"""
Main file to run all examples contained in this source distribution
"""
import os, sys
import subprocess

all_examples = [
    'trusses',
    'beams',
    'frames',
    'plates',
    'solids',
    'mixed_structures',
]

if __name__ == "__main__":

    for example_folder in all_examples:
        try:
            os.chdir(example_folder)
        except:
            continue

        print(f"Running examples in: {example_folder}")

        try:
            subprocess.call([sys.executable, "./runall.py"])
        except:
            print("... no runall.py script found in this folder")

        os.chdir('..')



import numpy as np

from .FrameTransformation import *

class BeamTransformation(FrameTransformation):
    r"""
    This is an alias for :py:meth:`FrameTransformation`
    """

    def __init__(self, dir1=None, dir2=None, axis=None):
        super().__init__(dir1, dir2, axis)



"""
==============================================
4-node quadrilateral - small displacement
==============================================
Hellinger-Reissner formulation for a quad with bi-linear
displacement interpolation.

.. code::

         1
      x     y
        x*y


"""
import numpy as np
from copy import deepcopy

from ..Element import *
from ...domain.Node import *
from ...utilities import QuadIntegration, QuadShapes

class HRQuad(Element):
    """
    class: representing a plane 4-node quadrilateral

    This element works as 2D plate, loaded in-plane, and as a 3D membrane element.

    * For 2D plate behavior, define nodes as two-dimensional nodes
    * For 3D membrane behavior, define nodes as three-dimensional nodes
    """

    def __init__(self, node0, node1, node2, node3, material, label=None):
        super(HRQuad, self).__init__((node0, node1, node2, node3), material, label=label)
        self.element_type = DrawElement.QUAD
        self.createFaces()

        if node0.getPos().size == 3:
            dof_list = ('ux','uy','uz')
            ndof = 3
        elif node0.getPos().size == 2:
            dof_list = ('ux','uy')
            ndof = 2
        else:
            raise TypeError("dimension of nodes must be 2 or 3")

        self._requestDofs(dof_list)

        self.distributed_load = [0.0, 0.0, 0.0, 0.0]   # face loads along the perimeter of the element
        self.force    = 0.0
        self.Forces   = [ np.zeros(ndof) for k in range(len(self.nodes)) ]
        self.Kt       = [ [ np.zeros(ndof) for k in range(len(self.nodes)) ] for m in range(len(self.nodes)) ]
        self.ndof = ndof

        self.material  = []   # one material object per integration point
        self.stress    = []   # hold gauss-point stress (1st Piola-Kirchhoff stress)
        self.J         = []   # jacobian at integration point

        self.Grad      = []   # derivative of shape functions with respect to global coords

        X  = np.array([ node.getPos() for node in self.nodes ])

        # initialization step
        integrator = QuadIntegration(order=2)
        xis, wis = integrator.parameters()

        gpt = 0

        interpolation = QuadShapes()

        for xi, wi in zip(xis, wis):

            dphi_ds = interpolation.shape(1, *xi, n=(1,0))
            dphi_dt = interpolation.shape(1, *xi, n=(0,1))
            Grad = np.vstack((dphi_ds, dphi_dt))

            # reference configuration
            # -------------------------

            #
            # $ D\Phi_0 $
            #
            DPhi0 = (Grad @ X).T
            self.J.append( np.linalg.det(DPhi0) )  # material model already includes thickness

            # dual base (contra-variant)
            self.Grad.append( np.linalg.inv(DPhi0).T @ Grad)

            self.material.append(deepcopy(material))
            self.stress.append({})
            gpt += 1



    def __str__(self):
        s = super(HRQuad, self).__str__()
        for igpt, material in enumerate(self.material):
            s += "\n    strain ({}): xx={xx:.3e} yy={yy:.3e} xy={xy:.3e} zz={zz:.3e}".format(igpt,**material.getStrain())
            s += "\n    stress ({}): xx={xx:.3e} yy={yy:.3e} xy={xy:.3e} zz={zz:.3e}".format(igpt,**material.getStress())
        if np.array(self.distributed_load).any():
            s += "\n    element forces added to node:"
            for i, P in enumerate(self.Loads):
                Pi = np.array(P) * self.loadfactor
                s += "\n        {}: {}".format(self.nodes[i].getID(), Pi)
        return s

    def setSurfaceLoad(self, face, pn, ps=0):
        """
        .. list-table::
            :header-rows: 1

            * - face ID
              - nodes defining that face
            * - 0
              - :py:obj:`node 0` to :py:obj:`node 1`
            * - 1
              - :py:obj:`node 1` to :py:obj:`node 2`
            * - 2
              - :py:obj:`node 2` to :py:obj:`node 3`
            * - 3
              - :py:obj:`node 3` to :py:obj:`node 0`


        :param face: face ID for the laoded face
        :param pn: magnitude of distributed normal load per area. Tension on a surface is positive.
        :param ps: magnitude of distributed shear load per area. Positive direction is defined as shown in the above table.
        """
        if face >= 0 and face <= 3:
            self.faces[face].setLoad(pn, ps)

    def resetLoads(self):
        super(HRQuad, self).resetLoads()

    def updateState(self):

        # initialization step
        integrator = QuadIntegration(order=2)

        nnds = len(self.nodes)
        ndof = self.ndof       # mechanical element

        self.Forces = [ np.zeros(ndof) for k in range(nnds) ]
        R  = [ np.zeros((ndof,ndof)) for i in range(nnds) ]
        Kt = [ [ np.zeros((ndof,ndof)) for i in range(nnds) ] for j in range(nnds) ]

        # create array of deformed nodal coordinates
        xt = np.array([ node.getDeformedPos() for node in self.nodes ])

        gpt = 0

        # interpolation = QuadShapes()   # we are doing that and the isoparametric transformation in the constructor

        xis, wis = integrator.parameters()

        for xi, wi in zip(xis, wis):

            # dphi_ds = interpolation.shape(1, *xi, n=(1,0))
            # dphi_dt = interpolation.shape(1, *xi, n=(0,1))
            # Grad = np.vstack((dphi_ds, dphi_dt))

            # reference configuration
            # -------------------------

            Grad = self.Grad[gpt]  # pre-computed in __init__
            wi *= self.J[gpt]      # J includes the thickness of the plate

            # spatial configuration
            # -------------------------

            # deformation gradient
            F = (Grad @ xt).T

            # compute small strain tensor
            eps = 0.5 * ( F + F.T ) - np.eye(self.ndof)

            # update the material state
            strain = {'xx':eps[0,0], 'yy':eps[1,1], 'xy':eps[0,1]+eps[1,0]}

            self.material[gpt].setStrain(strain)

            # 2nd Piola-Kirchhoff stress
            stress = self.material[gpt].getStress()

            S = np.array( [[stress['xx'],stress['xy']],[stress['xy'],stress['yy']]] )

            # store stress for reporting
            self.stress[gpt] = {'xx':S[0,0], 'xy':S[0,1], 'yx':S[1,0], 'yy':S[1,1]}

            Ones = np.eye(self.ndof)
            # compute kinematic matrices
            BI = [ np.array([ Grad[0,K]*Ones[:,0],                         # the XX component
                              Grad[1,K]*Ones[:,1],                         # the YY component
                              Grad[0,K]*Ones[:,1] + Grad[1,K]*Ones[:,0] ]) # the XY component is XY + YX ("gammaXY = 2 epsXY")
                   for K in range(nnds) ]

            # internal forces
            for i, force in enumerate(self.Forces):
                force += S @ Grad[:,i] * wi  # wi already includes J

            # tangent stiffness
            Ct = self.material[gpt].getStiffness() * wi

            for I, Bi in enumerate(BI):
                GCti = Bi.T @ Ct
                for J, Bj in enumerate(BI):
                    Kt[I][J] += GCti @ Bj

            # on to the next integration point
            gpt += 1

        self.Kt = Kt

    def computeSurfaceLoads(self):
        """
        compute surface loads using faces

        This method should be called during :py:meth:`updateState()` by every
        element supporting surface loads

        """
        self.Loads = [ np.zeros_like(self.Forces[I]) for I in range(len(self.nodes)) ]

        for I, face in enumerate(self.faces):
            loads = face.computeNodalForces()

            # indexing
            J = I+1
            if J>3:
                J -= 4

            # add to element load vectors
            self.Loads[I] += loads[0]
            self.Loads[J] += loads[1]
            if loads.shape[0]>2:
                numNodes = len(self.nodes)
                if numNodes == 8 or numNodes == 9:
                    K = I + 4
                else:
                    msg = "Force data provided from Face2D inconsistent with element data"
                    raise TypeError(msg)

                self.Loads[K] += loads[2]

    def getStress(self):
        return self.Stress






r"""
==============================================
A 1d linear spring element
==============================================

The spring is defined as the horizontal connection between
two nodes :py:obj:`ndi` and :py:obj:`ndj`.

The mechanics of the element is described by length change

.. math::

   \delta = U_{j} - U_{i}

and internal force

.. math::

   f = c\:\delta

where :math:`c` is the spring stiffness in force per length change.


"""
import numpy as np
from ..Element import Element
from ...materials import Material

class Spring(Element):
    """
    A 1d linear elastic spring element.

    This element only uses the x-coordinate and the displacement in x-direction ('ux').
    """

    def __init__(self, ndi, ndj, c=1, label=None):
        """
        :param ndi:  node object
        :param ndj:  node object
        :patam c:    spring stiffness (force/length change)
        """
        super(Spring, self).__init__((ndi, ndj), Material(params={'E':1.0, 'A':c}), label=label)
        self.element_type = Element.LINE
        self.c = c

        self._requestDofs(['ux'])

        # initialization
        self.force  = 0.0           # internal force
        self.delta  = 0.0           # length change
        self.Forces = [0.0,0.0]     # nodal forces
        self.Kt = [[c, -c],[-c, c]] # spring stiffness matrix

    def __str__(self):
        """
        nice format for printing this spring
        """
        s = "Spring {}: {} to {} with c={}".format(self.getID(), self.nodes[0].getID(), self.nodes[1].getID(), self.c)
        s += f"\n    length change:  delta = {self.delta} "
        s += f"\n    internal force: force = {self.force}"
        return s

    def __repr__(self):
        """
        nice format for debugging
        """
        s = "Spring({}, {}, c={})".format(self.getID(), self.nodes[0].getID(), self.nodes[1].getID(), self.c)
        return s

    def updateState(self):
        """
        called to compute internal force and current nodal reactions for equilibrium test

        :return: no return values needed (inherited from Element class)
        """
        Ui = self.getDisp(0)[0]  # this functions ALWAYS returns a list. We need only the first element
        Uj = self.getDisp(1)[0]  # this functions ALWAYS returns a list. We need only the first element

        # length change (elongation positive)
        self.delta = Uj - Ui

        # internal force (tension positive)
        self.force = self.c * self.delta

        # nodal forces (global coordinates)
        self.Forces = [-self.force, self.force]

    def getInternalForce(self, variable=''):
        """
        Helper function used by the plotting functions.

        :return: :py:obj:`(s,f)` , two lists containing start and end values of normalized position and internal force,
                respectively, at both the start and the end point of the line of action.
        """
        s = np.array( [0.0, 1.0] )
        f = np.array( [-self.Forces[0], self.Forces[1]] )
        return (s, f)




"""
Example: plate03

Patch test for the LinearTriangle (Constant Strain Triangle)

"""
from ...examples.Example import *

from ...domain import *
from ...solver.NewtonRaphsonSolver import *
from ...elements.LinearTriangle import *
from ...materials.PlaneStress import *


class ExamplePlate03(Example):

    def docString(self):
        s = """
    ## Patch test for triangular plate under in-plane loading

    The patch test is an empirical minimum test which every finite element has to pass to ensure convergence with mesh refinement.

    It consists of a problem for which a known homogeneous solution exists.  For plates, we commonly use a rectangular plate subject to homogenous edge loading, e.g., constant tension in the x-direction, or constant shear, etc.

    The mesh must contain distorted elements and at least one element not attached to any node on the boundary.
        
    Author: Peter Mackenzie-Helnwein 
    """
        return s

    def problem(self):
        # ========== setting mesh parameters ==============

        N = 8         # number of elements in the mesh
        Lx = 100.0    # length of plate in the x-direction
        Ly =  80.0    # length of plate in the y-direction

        # ========== setting material parameters ==============

        params = dict(
            E  = 20000.,    # Young's modulus
            nu = 0.250,     # Poisson's ratio
            t  = 1.00       # thickness of the plate
        )

        # ========== setting load parameters ==============

        px  = 10.0         # uniform load normal to x=const
        py  =  0.0         # uniform load normal to y=const
        pxy =  0.0         # uniform shear load on x=const and y=const

        # ========== setting analysis parameters ==============

        target_load_level = 1.00     # reference load
        max_steps = 2                # number of load steps: 2 -> [0.0, 1.0]

        # define a list of target load levels
        load_levels = np.linspace(0, target_load_level, max_steps)

        #
        # ==== Build the system model ====
        #

        model = System()
        model.setSolver(NewtonRaphsonSolver())

        # create nodes

        nodes = (
            Node(0.0*Lx, 0.0*Ly),  # nd 0
            Node(0.2*Lx, 0.0*Ly),  # nd 1
            Node(0.5*Lx, 0.0*Ly),  # nd 2
            Node(0.7*Lx, 0.0*Ly),  # nd 3
            Node(1.0*Lx, 0.0*Ly),  # nd 4
            #
            Node(0.0*Lx, 0.2*Ly),  # nd 5
            Node(0.15*Lx,0.3*Ly),  # nd 6
            Node(0.5*Lx, 0.2*Ly),  # nd 7
            Node(0.8*Lx, 0.3*Ly),  # nd 8
            Node(1.0*Lx, 0.2*Ly),  # nd 9
            #
            Node(0.0*Lx, 0.6*Ly),  # nd 10
            Node(0.2*Lx, 0.5*Ly),  # nd 11
            Node(0.7*Lx, 0.7*Ly),  # nd 12
            Node(0.9*Lx, 0.6*Ly),  # nd 13
            Node(1.0*Lx, 0.8*Ly),  # nd 14
            #
            Node(0.0*Lx, 1.0*Ly),  # nd 15
            Node(0.3*Lx, 1.0*Ly),  # nd 16
            Node(0.55*Lx,1.0*Ly),  # nd 17
            Node(0.8*Lx, 1.0*Ly),  # nd 18
            Node(1.0*Lx, 1.0*Ly),  # nd 19
        )

        elements = (
            LinearTriangle(nodes[0],nodes[1],nodes[5],PlaneStress(params)),  # elem 0
            LinearTriangle(nodes[1],nodes[2],nodes[6],PlaneStress(params)),  # elem 1
            LinearTriangle(nodes[2],nodes[3],nodes[7],PlaneStress(params)),  # elem 2
            LinearTriangle(nodes[3],nodes[4],nodes[8],PlaneStress(params)),  # elem 3
            #
            LinearTriangle(nodes[6],nodes[5],nodes[1],PlaneStress(params)),  # elem 4
            LinearTriangle(nodes[7],nodes[6],nodes[2],PlaneStress(params)),  # elem 5
            LinearTriangle(nodes[8],nodes[7],nodes[3],PlaneStress(params)),  # elem 6
            LinearTriangle(nodes[9],nodes[8],nodes[4],PlaneStress(params)),  # elem 7
            #
            LinearTriangle(nodes[5],nodes[6],nodes[10],PlaneStress(params)),  # elem 8
            LinearTriangle(nodes[6],nodes[7],nodes[11],PlaneStress(params)),  # elem 9
            LinearTriangle(nodes[7],nodes[8],nodes[12],PlaneStress(params)),  # elem 10
            LinearTriangle(nodes[8],nodes[9],nodes[13],PlaneStress(params)),  # elem 11
            #
            LinearTriangle(nodes[11],nodes[10],nodes[6],PlaneStress(params)),  # elem 12
            LinearTriangle(nodes[12],nodes[11],nodes[7],PlaneStress(params)),  # elem 13
            LinearTriangle(nodes[13],nodes[12],nodes[8],PlaneStress(params)),  # elem 14
            LinearTriangle(nodes[14],nodes[13],nodes[9],PlaneStress(params)),  # elem 15
            #
            LinearTriangle(nodes[10],nodes[11],nodes[15],PlaneStress(params)),  # elem 16
            LinearTriangle(nodes[11],nodes[12],nodes[16],PlaneStress(params)),  # elem 17
            LinearTriangle(nodes[12],nodes[13],nodes[17],PlaneStress(params)),  # elem 18
            LinearTriangle(nodes[13],nodes[14],nodes[18],PlaneStress(params)),  # elem 19
            #
            LinearTriangle(nodes[16],nodes[15],nodes[11],PlaneStress(params)),  # elem 20
            LinearTriangle(nodes[17],nodes[16],nodes[12],PlaneStress(params)),  # elem 21
            LinearTriangle(nodes[18],nodes[17],nodes[13],PlaneStress(params)),  # elem 22
            LinearTriangle(nodes[19],nodes[18],nodes[14],PlaneStress(params)),  # elem 23
        )

        model.addNode(*nodes)
        model.addElement(*elements)

        # define support(s)

        fix_x = (0,)
        fix_y = (0,4)

        for idx in fix_x:
            nodes[idx].fixDOF('ux')    # horizontal support left end
        for idx in fix_y:
            nodes[idx].fixDOF('uy')          # vertical support right end

        # ==== complete the reference load ====

        # surface loads on the left side
        elements[ 0].setSurfaceLoad(2,px)
        elements[8].setSurfaceLoad(2,px)
        elements[16].setSurfaceLoad(2,px)

        # surface loads on the right side
        elements[ 7].setSurfaceLoad(2,px)
        elements[15].setSurfaceLoad(2,px)
        elements[23].setSurfaceLoad(2,px)

        # these are only nodal forces as part of the reference load
        # .. load only the upper node
        #print(model)

        model.report()

        model.plot(factor=10., title="undeformed system", filename="plate03_undeformed.png")

        model.setLoadFactor(10.0)
        model.solve()

        model.report()

        model.plot(factor=10., filename="plate03_deformed.png")




"""
Example: frame03

Buckling of a building frame

modeled using a 2D frame element

.. list-table:: setting given parameters

    * - N  = 2
      - number of elements
    * - L  = 100.0
      - column length
    * - EA = 2000000.0
      - axial stiffness
    * - EI = 21000.0
      - flexural stiffness
    * - w  = 0.1
      - applied lateral load


"""
from ...examples.Example import *

from ...domain import *
from ...solver.NewtonRaphsonSolver import *
from ...elements.Frame2D import *
from ...materials.ElasticSection import *


class ExampleFrame03(Example):

    def docString(self):
        s = """
    Buckling of a building frame

    degrees of freedom:
        ux ... horizontal displacement
        uy ... vertical displacement
        rz ... rotation, theta
        
    Author: Peter Mackenzie-Helnwein 
    """
        return s

    def problem(self):
        # initialize a system model

        N  = 8     # number of elements

        B = 240.
        H = 200.

        E  = 29000.0
        A = 20.0
        I = 10.0

        w = 1.0
        load_at_nodes_only = False # set to True to apply equivalent nodal forces and moments

        Ph = 0.01      # additional horizontal load per floor
        Ph = 0.10      # additional horizontal load per floor
        Ph = 1.00      # additional horizontal load per floor
        Ph = 0.00      # additional horizontal load per floor

        # ========== setting global parameters ==============

        target_load_level = 27
        max_steps = 10
        load_levels = np.linspace(0, target_load_level, max_steps)

        # ========= build your structural model =============

        model = System()
        model.setSolver(NewtonRaphsonSolver())

        x0 = 0.0
        x1 = B / 3
        x2 = 2 * B / 3
        x3 = B

        y0 = 0.0
        y1 = H / 4
        y2 = 2 * H / 4
        y3 = 3 * H / 4
        y4 = H

        X10 = Node(x0, y0)
        X11 = Node(x0, y1)

        X20 = Node(x1, y0)
        X21 = Node(x1, y1)

        model.addNode(X10,X11)
        model.addNode(X20,X21)

        # columns

        params = {'E': E, 'A': A, 'I': I}

        C11 = Frame2D(X10, X11, ElasticSection(params))
        C21 = Frame2D(X20, X21, ElasticSection(params))

        model.addElement(C11,C21)

        # floors

        params = {'E': E, 'A': A, 'I': 8*I}

        F11 = Frame2D(X11, X21, ElasticSection(params))

        model.addElement(F11)

        # fixities
        X10.fixDOF('ux','uy','rz')   # fixed
        X20.fixDOF('ux','uy','rz')   # fixed

        # reference load
        #Pcr = np.pi**2 * EI / L**2
        model.resetLoad()            # size load vector and initialize
        #model.addLoad(Xn, -Pcr, dof=0) # add a horizontal force (first dof only) ; remember C-style indexing: 0,1,...,(n-1)

        if load_at_nodes_only:

            # floor loading as nodal loads ...

            Pe = w * B/3
            Mi = w * (B/3)**2 /12

            X11.addLoad([-Pe/2., -Mi],['uy','rz'])
            X21.addLoad([-Pe/2.,  0.],['uy','rz'])

        else:

            # floor loading as distributed loads ...

            F11.setDistLoad(-w)


        # wind load ...

        X11.addLoad([Ph],['ux'])   # horizontal load


        # show model information
        print(model)

        print("\n==== perform the analysis ===\n")

        # * apply the load in multiple smaller load steps

        # set up data recorder
        model.initRecorder()
        model.trackStability(True)

        # initialize the analysis:
        model.resetDisplacements()   # set U to all zeros
        model.setLoadFactor(0.0)     # define a known equilibrium solution

        model.startRecorder()

        detKt   = []
        lambdas = []

        # solve for all load_levels
        for loadfactor in load_levels:

            # define node X2 as the controled node; downward direction is prescribed:
            model.setLoadFactor(loadfactor)
            model.solve(verbose=True)

            # stability check
            lambdas.append(model.loadfactor)
            detKt.append(model.solver.checkStability())

            # report results
            print('+')
            #model.report()

            print("\n=== next load level ===\n")


        #
        # ==== create some nice plots ===
        #

        model.report()

        print(lambdas)
        print(detKt)

        plt.plot(lambdas,detKt,'--*r')
        plt.grid(True)
        plt.xlabel('Load factor, $ \\lambda $')
        plt.ylabel("Stability index, $ {det}\: {\\bf K}_t $")
        plt.show()

        model.plot(factor=1.0, filename="frame3_deformed.png")

        model.beamValuePlot("F", filename="frame3_force.png")
        model.beamValuePlot("V", filename="frame3_shear.png")
        model.beamValuePlot("M", filename="frame3_moment.png")

        model.plotBucklingMode(factor=25., filename="frame1_buckling_mode0.png")




import numpy as np

from .Material import *

class VonMises(Material):
    """
    class: representing a 3d von Mises Material

    Assumptions are isotropic linear elastic-hardening plastic behavior.

    """

    def __init__(self, params={'E':1.0, 'nu':0.0, 'fy':1.0e30}):
        super().__init__(params = params)

        self._type = self.PLANESTRESS | self.PLASTIC | self.HARDENING

        # isotropic hardening parameter
        if 'H' not in self.parameters:
            self.parameters['H'] = 0.0

        # kinematic hardening parameter
        if 'K' not in self.parameters:
            self.parameters['K'] = 0.0

        # initialize strain
        self.plastic_strain = np.zeros(6)
        self.alpha          = 0.0
        self.beta           = np.zeros(6)

        self.setStrain({'xx':0.0, 'yy':0.0, 'zz':0.0, 'xy':0.0, 'yz':0.0, 'zx':0.0})

    def getThickness(self):
        raise NotImplementedError('Thickness has no meaning in a 3d model')

    def updateState(self, eps):
        """
        update state for a user provided axial strain value

        :param eps:  strain or strain tensor
        :return: n/a
        """

        # update stress state
        E  = self.parameters['E']
        nu = self.parameters['nu']
        fy = self.parameters['fy']
        H  = self.parameters['H']   # kinematic hardening parameter
        K  = self.parameters['K']   # isotropic hardening parameter
        twoG  = E / (1 + nu)

        Idev = np.array([ [2./3., -1./3., -1./3.,  0.,  0.,  0.],
                          [2./3., -1./3., -1./3.,  0.,  0.,  0.],
                          [2./3., -1./3., -1./3.,  0.,  0.,  0.],
                          [   0.,     0.,     0., 0.5,  0.,  0.],
                          [   0.,     0.,     0.,  0., 0.5,  0.],
                          [   0.,     0.,     0.,  0.,  0., 0.5] ])
        Ce_inv = (1 / E) * np.array([ [1., -nu, -nu, 0., 0., 0.],
                                    [-nu, 1., -nu, 0., 0., 0.],
                                    [-nu, -nu, 1., 0., 0., 0.],
                                    [ 0., 0., 0.,  2. * (1. + nu), 0., 0.],
                                    [ 0., 0., 0.,  0., 2. * (1. + nu), 0.],
                                    [ 0., 0., 0.,  0., 0., 2. * (1. + nu)] ])
        Ce = np.linalg.inv( Ce_inv )

        # epsilon_{n+1}
        eps1 = np.array([self.strain['xx'],self.strain['yy'],self.strain['zz'],
                        self.strain['xy'],self.strain['yz'],self.strain['zx']])

        # set trial state
        self.plastic_strain1 = self.plastic_strain.copy()
        self.alpha1          = self.alpha
        self.beta1           = self.beta.copy()

        self.Et = Ce

        # elastic predictor (trial state)
        self.sig = self.Et @ ( eps1 - self.plastic_strain )
        q_iso = - np.sqrt(2./3.) * K * self.alpha1
        q_kin = - H * self.beta1

        # check yield condition
        one = np.array( [1.,1.,1.,0.,0.,0.] )
        I1 = self.sig @ one
        eta = self.sig - (I1/3.) * one + q_kin
        norm_eta = np.sqrt( eta @ eta )

        f = norm_eta + np.sqrt(2./3.) * (q_iso - fy)

        # plastic corrector as needed
        if f >= 0.0:
            # loop
            r = eta / norm_eta
            Cxr = twoG * r  # Ce . r = 2G r   since tr(r) == 0
            gamma = f / (twoG + K + (2/3)*H)

            # update state variables
            self.alpha1 += np.sqrt(2/3) * gamma
            self.beta1  += gamma * r

            rstrain = r.copy()
            rstrain[3] *= 2  # gamma_xy^p = 2 eps_xy^p
            rstrain[4] *= 2  # gamma_yz^p = 2 eps_yz^p
            rstrain[5] *= 2  # gamma_zx^p = 2 eps_zx^p
            self.plastic_strain1 += gamma * rstrain

            self.sig = Ce @ ( self.strain - self.plastic_strain1 )

            Cep  = Ce - twoG*twoG*gamma/norm_eta * Idev
            Cep -= twoG*twoG/(twoG + K + (2/3)*H) * (1 - f/norm_eta) * np.outer(r,r)

            self.Et  = Cep

    def getStrain(self):
        return self.sig / self.parameters['E'] + self.plastic_strain

    def converged(self):
        # update state now that the global analysis has converged
        self.plastic_strain = self.plastic_strain1
        self.alpha          = self.alpha1
        self.beta           = self.beta1


if __name__ == "__main__":

    import matplotlib.pyplot as plt

    # testing the Node class
    mat = VonMises(params={'E':100.0, 'nu':0.0, 'fy':1.0})

    eps = 0.02 * np.sin( np.linspace(0, 2.*np.pi, 100) )

    sig = []
    Et = []
    for strain in eps:
        # update material strain state
        mat.setStrain(strain)
        #collect stress response
        sig.append(mat.getStress())
        #collect material tangent stiffness response
        Et.append(mat.getStiffness())

    fig, (ax1,ax2) = plt.subplots(2,1)

    ax1.plot(eps, sig, '-r', label='stress')
    ax1.grid(True)
    ax1.set_xlabel('strain $\\varepsilon$')
    ax1.set_ylabel('stress $\\sigma$')
    ax1.legend()

    ax2.plot(eps, Et, '-.b', label='tangent modulus')
    ax2.grid(True)
    ax2.set_xlabel('strain $\\varepsilon$')
    ax2.set_ylabel('tangent modulus $E_t$')
    ax2.legend()

    plt.show()



from .Recorder import *
from .Record import *

class ElementRecorder(Recorder):
    """

    """

    def __init__(self, **kwargs):
        super(ElementRecorder, self).__init__(**kwargs)

        if 'label' in kwargs:
            template = kwargs['label'] + ":{}"
        else:
            template = "elem:{}"

        for var in self.elemvars:
            lbl = template.format(var)
            if var == 'stress' or var == 'strain':
                for key in ('xx','yy','zz','xy','yz','zx',):
                    lbl += f":{key}"
                    reference = f"{var}:{key}"
                    self.data[reference] = Record(key=reference, label=lbl)
            else:
                self.data[var] = Record(key=var, label=lbl)

            print(lbl)




import numpy as np
from .Integration import Integration

class TetraIntegration(Integration):
    """
    Provides integration points and weights for tetrahedral domains
    """

    def __init__(self, order=2):
        super(TetraIntegration, self).__init__(order=order, dimension=1)
        self.init_points_and_weights()

    def init_points_and_weights(self):

        if self.order <= 1:
            xi = [ ( 0.25, 0.25, 0.25, 0.25 ) ]
            w  = [ 1/6. ]
        elif self.order == 2:
            a = ( 5. + 3.*np.sqrt( 5. )) / 20.
            b = ( 5. - np.sqrt( 5. )) / 20.
            wi = 1./24.
            xi = [ (a,b,b,b), (b,a,b,b), (b,b,a,b), (b,b,b,a) ]
            w  = [ wi, wi, wi, wi]
        elif self.order == 3:
            a = 0.500
            b = 1./6.
            wi = 9./120.
            xi = [ ( 0.25, 0.25, 0.25, 0.25 ), (a,b,b,b), (b,a,b,b), (b,b,a,b), (b,b,b,a) ]
            w  = [ -2./15., wi, wi, wi, wi ]
        else:
            a = 11./14.
            b = 1./14.
            wi = 343./45000.
            xi = [ ( 0.25, 0.25, 0.25, 0.25 ), (a,b,b,b), (b,a,b,b), (b,b,a,b), (b,b,b,a) ]
            w  = [ -74./5625., wi, wi, wi, wi ]
            a = 0.25 * ( 1. + np.sqrt( 5./14. ))
            b = 0.25 * ( 1. - np.sqrt( 5./14. ))
            wi = 56. / 2250.
            xi += [ (a,a,b,b), (a,b,a,b), (a,b,b,a), (b,a,a,b), (b,a,b,a), (b,b,a,a) ]
            w  += [ wi, wi, wi, wi, wi, wi ]

        self.xi = xi
        self.w  =  w



"""
Example
"""
from ...examples.Example import *

from ...domain.System import *
from ...domain.Node import *
from ...elements.Truss import *
from ...materials.FiberMaterial import *


class ExampleTruss03(Example):

    def docString(self):
        s = """
        Simple triangular truss. 
        
        This example is structurally identical to ExampleTruss01 but utilizes 
        the alternative input style for adding Nodes and Elements to the model.
        
        The system is statically determined and allows for easy validation of 
        calculated deformation, reactions and internal forces.
        
        Author: Peter Mackenzie-Helnwein 
        """
        return s

    def problem(self):
        # initialize a system model
        B = 6.0 * 12
        H = 3.0 * 12
        params = {'E': 10., 'A': 1., 'nu': 0.0, 'fy': 1.e30}

        model = System()

        # create nodes
        nd0 = Node(0.0, 0.0)
        nd1 = Node(  B, 0.0)
        nd2 = Node(0.5*B, H)

        model += nd0
        model += nd1
        model += nd2

        # create elements
        model += Truss(nd0, nd1, FiberMaterial(params))  # bottom 1
        model += Truss(nd0, nd2, FiberMaterial(params))  # up right diag 1
        model += Truss(nd1, nd2, FiberMaterial(params))  # up left diag 1

        # define support(s)
        nd0.fixDOF('ux')    # horizontal support left end
        #nd0 //= 0
        nd0.fixDOF('uy')    # vertical support left end
        nd1.fixDOF('uy')    # vertical support right end

        # add loads
        # .. load only the upper nodes
        nd2.setLoad((0.0, -1.0), ('ux','uy'))

        # analyze the model
        model.solve()

        # write out report
        model.report()

        # create plots
        model.plot(factor=1., filename="truss03_deformed_a.png")

        # fix horizontal motion of node 1
        nd1.fixDOF('ux')

        # add loads: same load -- nothing to do

        # RE-analyze the model
        model.resetDisp()
        model.solve()

        # skip the report
        model.report()

        # create plots
        model.plot(factor=1., filename="truss03_deformed_b.png")




from .Integration import Integration

class QuadIntegration(Integration):
    r"""
    Provides integration points and weights for quadrilateral domains (2D)
    """

    def __init__(self, order=2):
        super(QuadIntegration, self).__init__(order=order, dimension=1)
        self.init_points_and_weights(nGP=(order + 2)//2)

    def init_points_and_weights(self, nGP):
        xi, w = self.gauss1D( nGP, biunit=True )
        self.xi = [ (xi[i], xi[j]) for j in range(nGP) for i in range(nGP) ]
        self.w  = [  w[i]* w[j] for j in range(nGP) for i in range(nGP) ]




from .Plotter import *

class Plotter3d(Plotter):
    """
    class: representing a Plotter object
    """

    def __init__(self):
        super(Plotter3d, self).__init__()

    def displacementPlot(self, file=None):
        """
        Create a deformed 3d system plot

        If **file** is given, store the plot to that file.
        Use proper file extensions to indicate the desired format (.png, .pdf)

        :param file: filename (str)
        """
        fig, axs = plt.subplots()

        # plot the undeformed lines
        for line in self.lines:
            vert0 = self.vertices[line[0]]
            vert1 = self.vertices[line[1]]
            x = [vert0[0], vert1[0]]
            y = [vert0[1], vert1[1]]
            z = [vert0[2], vert1[2]]
            axs.plot(x,y,z,'-k',lw=2)

        # plot the deformed lines
        if len(self.disp) == len(self.vertices):
            for line in self.lines:
                vert0 = self.vertices[line[0]].copy()
                vert1 = self.vertices[line[1]].copy()
                vert0 += self.disp[line[0]]
                vert1 += self.disp[line[1]]
                x = [vert0[0], vert1[0]]
                y = [vert0[1], vert1[1]]
                z = [vert0[2], vert1[2]]
                axs.plot(x,y,z,'-r',lw=3)

        if self.reactions != []:
            self.addForces(axs)

        #axs.set_aspect('equal')
        #axs.set_axis_off()

        plt.show()

    def valuePlot(self, deformed=False, file=None):
        """
        Create a plot using colors to identify magnitude of internal force.

        If **file** is given, store the plot to that file.
        Use proper file extensions to indicate the desired format (.png, .pdf)

        :param deformed: True | **False**
        :param file: filename (str)
        """
        fig, axs = plt.subplots()

        # plot the lines
        segments = []

        if len(self.disp) == len(self.vertices):
            for line in self.lines:
                vert0 = self.vertices[line[0]].copy()  # we need a copy since we will be modifying this in the lines below
                vert1 = self.vertices[line[1]].copy()  # we need a copy since we will be modifying this in the lines below
                if deformed:
                    vert0 += self.disp[line[0]]   # it's this += that modifies the vertices if we don't use a copy
                    vert1 += self.disp[line[1]]   # it's this += that modifies the vertices if we don't use a copy
                #x = [vert0[0], vert1[0]]
                #y = [vert0[1], vert1[1]]
                #axis.plot(x,y,'-r',lw=3)
                segments.append(np.array([vert0, vert1]))

        # Create a continuous norm to map from data points to colors
        lc = LineCollection(np.array(segments), cmap='rainbow')
        # Set the values used for colormapping
        lc.set_array(self.values)
        lc.set_linewidth(3)
        line = axs.add_collection(lc)
        fig.colorbar(line, ax=axs)

        if self.reactions != []:
            self.addForces(axs)

        axs.set_aspect('equal')
        axs.set_axis_off()

        plt.autoscale(enable=True, axis='x', tight=False)
        plt.autoscale(enable=True, axis='y', tight=False)
        plt.show()

    def addForces(self, axs):
        """
        add nodal forces to the plot shown in **axs**

        :param axs: axis on which to plot
        """
        if len(self.reactions) == len(self.vertices):
            Fx = []
            Fy = []
            X = []
            Y = []
            for (point, force) in zip(self.vertices, self.reactions):
                if np.linalg.norm(force) > 1.0e-3:
                    X.append(point[0])
                    Y.append(point[1])
                    Fx.append(-force[0])
                    Fy.append(-force[1])

            axs.quiver(X,Y, Fx, Fy, color='green')




if __name__ == "__main__":
    # testing the plotter
    plotter = Plotter3d()



from copy import deepcopy

import numpy as np
from collections import deque

from ..elements import Element
from ..recorder.Recorder import Recorder

class Node():
    r"""
    class: representing a single Node

    :param x0: Initial position (List)
    """
    COUNT = 0

    def __init__(self, x0, y0=None, z0=None):
        self.ID = Node.COUNT
        Node.COUNT += 1

        if isinstance(z0, (int, float)):
            self.pos = np.array([x0, y0, z0])
        elif isinstance(y0, (int, float)):
            self.pos = np.array([x0, y0])
        else:
            self.pos = np.array([x0])

        self.is_lead     = True   # is this a lead node?  Will be set to follower (is_lead = False) if tied
        self.lead        = self   # following yourself
        self.followers   = []     # list of following nodes

        self.disp          = None   # active current displacement vector
        self.disp_n        = None   # previously converged displacement vector
        self.loadfactor_n  = 0.0    # load factor for previously converged state
        self.disp_nn       = None   # two steps back converged displacement vector
        self.loadfactor_nn = 0.0    # load factor for two steps back converged state
        self.disp_mode     = None   # stored displacement representing a mode shape
        self.disp_pushed   = deque()   # stored displacement vector (see pushU() and popU())

        self.dofs        = {}
        self.ndofs       = 0
        self.start       = None
        self.elements    = []
        self._fixity     = []  # list for dof keys for fixed dofs
        self._setU       = {}  # prescribed displacement parameters u0 and u1: u[dof] = u0 + loadfactor*u1
        self.loads       = {}
        self._hasLoad    = False
        self._transform  = None    # nodal transformation object
        self.dof_maps    = {}      # dof_idx maps for attached elements

        self._resetGaussPointMap()
        self.setRecorder(None)

        self.setLoadFactor(1.0)

    def __str__(self):
        float_formatter = "{:.3f}".format
        np.set_printoptions(formatter={'float_kind': float_formatter})

        s  = "Node_{}:\n    x:    {}".format(self.ID, self.pos)
        if not self.is_lead:
            s +=  "\n    following {}".format(self.lead.getID())
        if self._transform:
            T = self._transform.getT()
            s += f"\n    local: x={T[:,0]}, y={T[:,1]}"
            if T.shape[1]>2:
                s += f", z={T[:,2]}"
        if self._fixity:
            s += f"\n    fix:  {self._fixity}"
        load = self.getLoad()
        if isinstance(load, np.ndarray) and not np.isclose(np.linalg.norm(load), 0.0):
            s += f"\n    P:    {load}"
        s += f"\n    u:    {self.disp}"

        float_formatter = "{:g}".format
        np.set_printoptions(formatter={'float_kind': float_formatter})

        return s

    def __repr__(self):
        return "Node_{}(x={}, u={})".format(self.ID, self.pos, self.disp)

    def getID(self):
        r"""
        :returns: the node ID (``str``)
        """
        return "Node_{}".format(self.ID)

    def getLead(self):
        r"""
        :returns: pointer to the **lead** node (``Node``)
        """
        return self.lead

    def getLeadID(self):
        r"""
        :returns: the node ID of the **lead** node (``str``)
        """
        return self.lead.getID()

    def request(self, dof_list, caller):
        r"""
        send list or individual dof code. Common codes:

        .. list-table::
            :header-rows: 1

            * - code
              - description
            * - **ux**
              - displacement in x-direction
            * - **uy**
              - displacement in y-direction
            * - **uz**
              - displacement in z-direction
            * - **rx**
              - rotation about x-axis
            * - **ry**
              - rotation about y-axis
            * - **rz**
              - rotation about z-axis


        :param dof_list:  list of dof-codes required by calling element
        :param caller:  pointer to calling element (usually sent as self)
        """
        if self.is_lead:
            dof_idx = []
            for dof in dof_list:
                if dof not in self.dofs:
                    self.dofs[dof] = self.ndofs
                    self.ndofs += 1
                dof_idx.append(self.dofs[dof])

            if caller not in self.elements:
                self.elements.append(caller)

            # remember the dof_idx map for this element for future interaction
            self.dof_maps[caller] = dof_idx

            # let any attached transformation know that the nodal dof-list has changed.
            if self._transform:
                self._transform.refreshMaps()

            return tuple(dof_idx)

        else:
            return self.lead.request(dof_list=dof_list, caller=caller)

    def fixDOF(self, *dofs):
        r"""
        provide a list of dof codes that shall be restrained

        :param dofs: fix the dof defined by key(s)

        See also:
           * :py:meth:`femedu.domain.Node.Node.setDOF`
           * :py:meth:`femedu.domain.Node.Node.request`
        """
        if self.is_lead:
            for dof in dofs:
                if isinstance(dof, str):
                    if dof not in self._fixity:
                        self._fixity.append(dof)
                elif isinstance(dof,list) or isinstance(dof,tuple):
                    for item in dof:
                        self.fixDOF(item)
                else:
                    raise TypeError
        else:
            self.lead.fixDOF(*dofs)

    def hasDOF(self, dof):
        """
        :param dof: a dof string ID
        :return: True(False) if dof in(not in) the node's dof-list
        """
        return (dof in self.dofs)

    def setDOF(self, dofs=[], values=[]):
        r"""
        alternative to the `fixDOF()` method that
        allows to set a prescribed value other than `0.0`
        for each degree of freedom.

        :param dofs: list of dofs, defined by key(s), for which displacement values shall be prescribed.
        :param values: list of prescribed values for the given list of dofs.
                       Values may be given as a float or as a list of two values.

                       * If only a single value is given, the respective dof will be set to that value.
                       * If a list is given, the dof will be set to :code:`value[0] + loadfactor*value[1]`

        See also:
           * :py:meth:`femedu.domain.Node.Node.fixDOF`
           * :py:meth:`femedu.domain.System.System.setLoadFactor`
           * :py:meth:`femedu.domain.Node.Node.request`
        """
        if (not dofs) or (len(dofs) != len(values)):
            msg = "setDOF requires matching non-empty lists of dofs and associated values"
            raise TypeError(msg)

        if self.is_lead:
            for dof, val in zip(dofs,values):
                self.fixDOF(dof)
                if isinstance(val,float) or isinstance(val,int):
                    self._setU[dof] = (val, 0.0)
                elif isinstance(val,list) or isinstance(val,tuple):
                    if len(val) == 1:
                        self._setU[dof] = (val[0], 0.0)
                    elif len(val) == 2:
                        self._setU[dof] = val
                    else:
                        raise TypeError(f"cannot interpret input values for {dof}")
                else:
                    raise TypeError
        else:
            self.lead.setDOF(dofs, values)

    def __floordiv__(self, other):
        r"""
        This is a short form for :py:meth:`Node.fixDOF(other)`

        :param other:
        :return: self
        """
        if self.is_lead:
            self.fixDOF(other)
            return self
        else:
            return self.lead.__floordiv__(other)

    def isFixed(self, dof):
        r"""

        :param dof: dof code as defined in :code:`request()`
        """
        if self.is_lead:
            return (dof in self._fixity)
        else:
            return self.lead.isFixed(dof)

    def areFixed(self):
        r"""
        To be used by the assembly routines.

        return a list of indices pointing to fixed dofs in this node.
        Indices are local to this node: :code:`0..num_dofs`
        """
        if self.is_lead:
            idx = []
            for i,dof in enumerate(self._fixity):
                idx.append(i)
            return np.array(idx, dtype=int)
        else:
            return self.lead.areFixed()


    def getFixedDofs(self):
        r"""
        :returns: a list of fixed dofs by dof-code strings.
        """
        if self.is_lead:
            return self._fixity[:]
        else:
            return self.lead.getFixedDofs()

    def setStart(self, startInt):
        if self.is_lead:
            self.start = startInt
        else:
            msg = "Illegal attempt to add {this_node} to the global system: {this_node} is a follower.".format(this_node=self.getID())
            raise TypeError(msg)

    def setDisp(self, U, dof_list=None, modeshape=False):
        r"""
        use for prescribed displacements

        :param U: list or tuple of prescribed values
        :param dof_list: list or tuple of dof-codes associated with values in U
        :param modeshape: set to True if U represents a mode shape (BOOL)
        """
        if self.is_lead:
            if isinstance(U,list) or isinstance(U,tuple):
                U = np.array(U)

            if dof_list:
                for ui, dof in zip(U, dof_list):
                    if dof in self.dofs:
                        if modeshape:
                            self.disp_mode[self.dofs[dof]] = ui
                        else:
                            self.disp[self.dofs[dof]] = ui
                    else:
                        msg = f"requested dof:{dof} not present at current node.  Available dofs are {self.dofs.keys()}"
                        raise TypeError(msg)
            else:
                if modeshape:
                    self.disp_mode = U
                else:
                    self.disp = U

        else:
            self.lead.setDisp(U, dof_list=dof_list, modeshape=modeshape)

    def _updateDisp(self, dU):
        r"""
        For internal use by solvers only

        :param dU: displacement correction from last iteration step.
        """
        if self.is_lead:

            if self._transform:
                self.disp += self.v2g(dU, self)
            else:
                self.disp += dU

        """
        Do not forward that call to the lead node or that increment will be duplicated.
        """

    def pushU(self):
        r"""
        Store the current displacement vector for later restore using :code:`popU()`.
        """
        self.disp_pushed.append({'U':self.disp.copy(), 'lam':self.loadfactor})
        if len(self.disp_pushed)>2:
            self.disp_pushed.popleft()

    def popU(self):
        r"""
        Restore a previously pushed displacement vector (using :code:`pushU()`).
        """
        if len(self.disp_pushed):
            state = self.disp_pushed.pop()
            self.disp       = state['U']
            self.loadfactor = state['lam']
        else:
            raise TypeError("no pushed displacement data available")

    def getDisp(self, dofs=None, caller=None, **kwargs):
        r"""
        return a vector (`nd.array`) of (generalized) displacements.

        If a :code:`caller` is given, the element-specific d.o.f.s will be returned as a sequence
        in the same order as given by the element's :code:`request()` call.

        If a :code:`dof_list` is given, d.o.f.s will be returned as the sequence given by that list.
        A zero value will be returned for all d.o.f.s that do not exist at this node.

        .. note::

            A single d.o.f., e.g., "ux", can be requested using :code:`getDisp(dofs=('ux',))`.
            Do not forget the :code:`,` to indicate the definition of a single-element tuple.

        :param caller: pointer to element
        :param dofs: tuple or list of d.o.f. keys
        :return: nodal displacement vector
        """
        if self.is_lead:
            if 'modeshape' in kwargs and kwargs['modeshape']:
                if not isinstance(self.disp_mode, np.ndarray):
                    self.disp_mode = np.zeros(self.ndofs)
                U = self.disp_mode
            else:
                if not isinstance(self.disp, np.ndarray):
                    self.disp    = np.zeros(self.ndofs)
                    self.disp_n  = np.zeros(self.ndofs)
                    self.disp_nn = np.zeros(self.ndofs)
                U = self.disp

            # so far, U is in global coordinates.
            # see if local coordinates were requested
            if 'local' in kwargs and kwargs['local'] == True:
                # check if we have a nodal transformation attached
                #if isinstance(self._transform, Transformation):
                if self._transform:
                    #
                    # see Element.getLoad() and Element.getForce() methods
                    #
                    U = self.v2l(U, self)

            # apply prescribed displacements
            for dof in self._setU:
                if dof in self.dofs:
                    # the index of dof in self.U
                    idx = self.dofs[dof]
                    # the prescribed displacement value
                    ubar = self._setU[dof][0] + self.loadfactor * self._setU[dof][1]
                    # set prescribed value in nodal U-vector
                    U[idx] = ubar

            if caller:
                # we know the calling element.
                # ... ignoring dofs and using dof list from element map

                if caller not in self.dof_maps:
                    msg = "caller not registered with this node"
                    raise TypeError(msg)

                idx = self.dof_maps[caller]
                ans = U[idx]
                return np.array(ans)

            else:
                # we do not know who is requesting displacements, so provide all requested or ALL if no dofs were specified.
                if dofs:
                    if isinstance(dofs,str):
                        dofs = [dofs]
                    ans = []
                    for dof in dofs:
                        if dof in self.dofs:
                            ans.append(U[self.dofs[dof]])
                        else:
                            ans.append(0.0)
                else:
                    ans = U

                return np.array(ans)

        else:
            return self.lead.getDisp(dofs=dofs, caller=caller, **kwargs)

    def getDeltaU(self, previous_step=False):
        r"""
        :return: delta u = (current u) - (last converged u)
        """
        if self.is_lead:
            if previous_step:
                dU = self.disp_n - self.disp_nn
            else:
                dU = self.disp - self.disp_n
            return dU
        else:
            return self.lead.getDeltaU(previous_step=previous_step)

    def getNormDeltaU2(self, previous_step=False):
        r"""
        :return: norm of delta u = (current u) - (last converged u)
        :return: 0.0 if node is a "follower"
        """
        if self.is_lead:
            dU = self.getDeltaU(previous_step)
            return dU @ dU
        else:
            return 0.0

    def getPos(self, caller=None, **kwargs):
        r"""
        :param caller: pointer to calling element
        :return: initial position vector
        """
        return self.pos

    def getDeformedPos(self, caller=None, factor=1.0, **kwargs):
        r"""
        Return deformed position :math:`{\bf x} = {\bf X} + f \: {\bf u}`

        If a caller is specified, the node will adjust the output to the d.o.f.s specified by the
        element's request() call. (called during initialization.)

        :param caller: pointer to the calling element.
        :param factor: deformation magnification factor, :math:`f`.
        :return: deformed position vector, :math:`{\bf x}`.
        """
        if self.is_lead:
            if self.pos.shape[0] == 1:
                my_dofs = ('ux',)
            elif self.pos.shape[0] == 2:
                my_dofs = ('ux','uy')
            elif self.pos.shape[0] == 3:
                my_dofs = ('ux','uy','uz')
            else:  # don't know what that would be
                msg = f"Don't know how to interprete position: {self.pos}"
                raise TypeError(msg)

            if 'modeshape' in kwargs and kwargs['modeshape']:
                if not isinstance(self.disp_mode, np.ndarray):
                    self.disp_mode = np.zeros(self.ndofs)

                return self.pos + factor * self.getDisp(caller=None, dofs=my_dofs, modeshape=1)
            else:
                if not isinstance(self.disp, np.ndarray):
                    self.disp    = np.zeros(self.ndofs)
                    self.disp_n  = np.zeros(self.ndofs)
                    self.disp_nn = np.zeros(self.ndofs)

                return self.pos + factor * self.getDisp(caller=None, dofs=my_dofs)

        else:
            if self.pos.shape[0] == 1:
                my_dofs = ('ux',)
            elif self.pos.shape[0] == 2:
                my_dofs = ('ux','uy')
            elif self.pos.shape[0] == 3:
                my_dofs = ('ux','uy','uz')
            else:  # don't know what that would be
                msg = f"Don't know how to interprete position: {self.pos}"
                raise TypeError(msg)

            #return self.lead.getDeformedPos(caller=caller, factor=factor, **kwargs)
            disp = self.lead.getDisp(dofs=my_dofs, caller=None, factor=factor, **kwargs)

            return self.pos + factor * disp

    def getIdx4Element(self, elem):
        r"""
        :return: an index list to nodal dofs associated with the attached **elem** within the global dof list
        """
        if self.is_lead:
            if elem in self.dof_maps:
                if self._transform:
                    # Transformation may expand the reach of an element to other components, hence, use all available dofs
                    return self.getIdx4DOFs()
                else:
                    # use the subset of dofs used by this element
                    return self.start + np.array(self.dof_maps[elem], dtype=int)
            else:
                msg = f"Element {elem} not in dof_map for node {self.ID}"
                raise TypeError(msg)
        else:
            return self.lead.getIdx4Element(elem)

    def getIdx4DOFs(self, dofs=[], local=False):
        r"""
        :param dofs:  a list of named dofs for which an index is requested.
                      The returned list will be in the same order as listed in **dofs**.
        :param local: by default return index to the global list of dofs.
                      Set :code:`local=True` to return index to this node's list of dofs.

        :return: an index list to this node's dofs
        """
        if self.is_lead:

            if not dofs:
                dofs = self.dofs.keys()

            idx = []
            for dof in dofs:
                if dof in self.dofs:
                    idx.append(self.dofs[dof])
                else:
                    msg = f"dof {dof} not present at node {self.ID}"
                    raise TypeError(msg)

            ans = np.array(idx, dtype=int)
            if not local:
                ans += self.start

            return ans

        else:
            return self.lead.getIdx4DOFs(dofs=dofs, local=local)

    def getLocalTransformationMap(self, dofs=[]):
        r"""
        :return: Transformation matrix
        """
        #if isinstance(self._transform, Transformation):
        if self._transform:
            T = self._transform.getT()
        else:
            T = None

        return T

    def addTransformation(self, T):
        r"""
        Attach a :py:meth:`femedu.domain.Transformation` object to this node.
        The transformation defines a local coordinate system.

        * If a transformation is given, all loads and prescribed displacements are assumed in that local coordinate system.
        * Furthermore, all nodal displacements, velocity, or acceleration will be reported in that local coordinate system.
        """
        from .Transformation import Transformation
        if isinstance(T, Transformation):
            T.registerClient(self) # register this Node with the transformation
            self._transform = T

    def addLoad(self, loads, dofs):
        r"""
        Nodes are applied as components energetically linked to any degree of freedom provided at this node.

        For example, in a 2D-problem with the x-axis to the right and the y-axis pointing up,
        applying a vertical downward force, P, and a counter-clockwise moment, M, at a node, xn:

        .. code::

            xn.addLoad([-P, M],['uy','rz'])

        :param loads:
        :type loads: list of floats
        :param dofs:
        :type dofs: list of dof-codes
        """
        if self.is_lead:
            # Check tuple type and if the dof exists (warn and continue)
            for (load, dof) in zip(loads, dofs):
                if dof in self.loads:
                    self.loads[dof] += load
                else:
                    self.loads[dof] = load
            self._hasLoad = True
        else:
            self.lead.addLoad(loads, dofs)

    def setLoad(self, loads, dofs):
        r"""
        :param loads: list of force components
        :param dofs:  associated list of DOFs to which respective loads are to be applied
        """
        if self.is_lead:
            # Check tuple type and if the dof exists (warn and continue)
            for (load, dof) in zip(loads, dofs):
                self.loads[dof] = load
            self._hasLoad = True
        else:
            self.lead.setLoad(loads, dofs)

    def resetLoad(self):
        """
        Resets the load vectors
        """
        if self.is_lead:
            self.loads = {}
            self._hasLoad = False
        else:
            self.lead.resetLoad()

    def getLoad(self, dof_list=None, apply_load_factor=False):
        r"""
        :param dof_list: list (or tuple) if dof-keys for which nodal loads are requested. Fill missing dofs by 0.0.
        :param apply_load_factor: defaults to False -> no factors applied.
        :returns: nodal load vector (ndarray)
        """
        if self.is_lead:
            if dof_list:
                force = np.zeros(len(dof_list))
                for dof in dof_list:
                    if (dof in self.loads) and (dof in self.dofs):
                            idx = dof_list.index(dof)
                            if apply_load_factor:
                                force[idx] = self.loads[dof] * self.loadfactor
                            else:
                                force[idx] = self.loads[dof]
            else:
                force = np.zeros(self.ndofs)
                for dof in self.loads:
                    if dof in self.dofs:
                        if apply_load_factor:
                            force[self.dofs[dof]] = self.loads[dof] * self.loadfactor
                        else:
                            force[self.dofs[dof]] = self.loads[dof]
        else:
            force = self.lead.getLoad(dof_list=dof_list, apply_load_factor=apply_load_factor)

        return force

    def hasLoad(self):
        r"""
        :returns: **True** if this node has **any** loads (bool)
        """
        if self.is_lead:
            return self._hasLoad
        else:
            return self.lead.hasLoad()

    def resetDisp(self):
        """
        Resets the displacement vector.
        """
        if self.is_lead:
            self.disp_nn = np.zeros(len(self.dofs))
            self.disp_n  = np.zeros(len(self.dofs))
            self.disp    = np.zeros(len(self.dofs))
        else:
            self.lead.resetDisp()

    def resetAll(self):
        """
        Resets load and displacement vectors.
        """
        self.resetDisp()
        self.resetLoad()

    def setLoadFactor(self, lam):
        r"""
        Set the target load factor to **lam**

        .. warning::

            This method should not be called by the user.
            **USE** :code:`System.setLoadFactor(lam)` instead!

        The entered load pattern is considered a reference load,
        to be multiplied by a scalar load factor, :math:`\lambda`.

        If no load factor is set explicitly, a factor of 1.0 is assumed, i.e., the
        entire entered load is applied in full.
        """
        self.loadfactor = lam

        if not self.is_lead:
            self.lead.setLoadFactor(lam)

    def isClose(self, x, TOL=1.0e-5):
        r"""
        :param x: :py:obj:`nd.array`
        :return: **True** if x is within **TOL** from this node.
        """
        pos = self.pos
        X = np.array(x)

        this_dim = pos.shape[0]
        x_dim = X.shape[0]

        if x_dim != this_dim:
            msg = f"target dimension ({x_dim}) incompatible with Node dimension ({this_dim})"
            raise TypeError(msg)

        is_close = (np.abs(pos - X) <= TOL).all()

        return is_close

    def distanceTo(self, x):
        r"""
        :param x: :py:obj:`nd.array`
        :return: scalar distance from x
        """
        pos = self.pos
        X = np.array(x)

        this_dim = pos.shape[0]
        x_dim = X.shape[0]

        if x_dim != this_dim:
            msg = f"target dimension ({x_dim}) incompatible with Node dimension ({this_dim})"
            raise TypeError(msg)

        dist = np.linalg.norm(pos - X)

        return dist

    def setRecorder(self, recorder):
        if isinstance(recorder, Recorder):
            self.recorder = recorder
        else:
            self.recorder = None

    def recordThisStep(self, load_level):
        r"""
        record current state of the system
        """
        if self.recorder and self.recorder.isActive():
            # data = {'lam':load_level}
            data = {}
            for var in self.recorder.getVariables():
                value = self.getDisp(var)[0]
                data[var] = value
            self.recorder.addData(data)

    def startRecorder(self):
        if self.recorder:
            self.recorder.enable()

    def pauseRecorder(self):
        if self.recorder:
            self.recorder.disable()

    def stopRecorder(self):
        if self.recorder:
            self.recorder.disable()

    def on_converged(self):
        r"""
        This method is called every time a solver signals a converged solution.
        """
        if self.is_lead:
            # rotate states (n)->(n-1) and current->(n)
            if isinstance(self.disp_n, np.ndarray):
                self.disp_nn = self.disp_n.copy()
            if isinstance(self.disp, np.ndarray):
                self.disp_n  = self.disp.copy()
            self.loadfactor_nn = self.loadfactor_n
            self.loadfactor_n  = self.loadfactor

        # do not duplicate this step for the load node !!

    def revert(self):
        pass

    def isLead(self):
        return self.is_lead

    def make_follower(self, lead):
        r"""
        this command tells the current node (**self**) to "follow" whatever the **lead** node is doing.

        :param lead: pointer to the lead node
        :type lead: Node
        """

        self.lead = lead

        if self == lead:
            self.is_lead = True
        else:
            self.is_lead = False
            lead.addFollower(self)

        # transfer element maps
        for elem in self.dof_maps:
            elem_dof_map = []
            for dof in self.dofs:
                idx = self.dofs[dof]
                if idx in self.dof_maps[elem]:
                    elem_dof_map.append(dof)
            lead.request(elem_dof_map, elem)

        # transfer nodal loads
        if self._hasLoad:
            self.lead.addLoad(self.loads.values(), self.loads.keys())
            self.loads = {}
            self._hasLoad = False

        # transfer fixities
        self.lead.fixDOF(self._fixity)

    def addFollower(self, follower):
        if follower in self.followers:
            msg = "{} is already following {}. Circular tie?".format(follower.getID(), self.getID())
            raise TypeError(msg)
        else:
            self.followers.append(follower)

    def setTrialState(self):
        r"""
        This will set a suitable trial position for the
        arc-length method to

        .. math::

           {\bf u}^{(0)}_{n+1} = 2{\bf u}^{(\infty)}_{n} - {\bf u}^{(\infty)}_{n-1}

        This should be used by a solution algorithm but not by regular
        user input.
        """
        self.disp       = 2.0 * self.disp_n - self.disp_nn
        self.loadfactor = 2.0 * self.loadfactor_n - self.loadfactor_nn

    #
    # Transformation methods
    #

    def hasTransform(self):
        return (self._transform != None)

    def v2l(self, U, caller=None):
        """
        transform a nodal vector from global to local coordinates

        This is commonly used by Elements to transform nodal forces to the Node's local
        system such that the :code:`Solver.assemble` method receives forces in local coordinates.

        :param U: vector to be transformed
        :param caller: pointer to the calling element
        :return: the transformed vector
        """
        if self.is_lead:

            if self._transform:
                # build a full length nodal vector
                if isinstance(caller, Element):
                    Unode = np.zeros(self.ndofs)                           # initialize to zeros
                    map = self.getIdx4DOFs(caller.getDofs(), local=True)   # get a map from element dogs to node dofs
                    Unode[map] = U                                         # fill in element vector to nodal vector
                else:
                    Unode = deepcopy(U)

                Ulocal = self._transform.v2l(Unode, self)
            else:
                Ulocal = deepcopy(U)

            return Ulocal

        else:
            return self.lead.v2l(U, caller=caller)

    def v2g(self, U, caller=None):
        """
        transform a nodal vector from local to global coordinates

        This is commonly used by the solver to transform nodal displacements from the Node's local to the global
        system such that the :code:`Node` class can store displacements internally in global coordinates.

        :param U: vector to be transformed
        :param caller: pointer to the calling element
        :return: the transformed vector
        """
        if self.is_lead:


            if self._transform:
                # build a full length nodal vector
                if isinstance(caller, Element):
                    Unode = np.zeros(self.ndofs)  # initialize to zeros
                    map = self.getIdx4DOFs(caller.getDofs(), local=True)  # get a map from element dogs to node dofs
                    Unode[map] = U  # fill in element vector to nodal vector
                else:
                    Unode = deepcopy(U)

                Uglobal = self._transform.v2g(Unode, self)
            else:
                Uglobal = deepcopy(U)

            return Uglobal

        else:
            return self.lead.v2g(U, caller=caller)

    def m2l(self, M, caller=None):
        """
        transform a nodal matrix from global to local coordinates

        This is commonly used by Elements to transform nodal forces to the Node's local
        system such that the :code:`Solver.assemble` method receives stiffness in local coordinates.

        :param M: vector to be transformed
        :param caller: pointer to the calling element
        :return: the transformed matrix
        """
        if self.is_lead:

            if self._transform:
                # build a full length nodal vector
                if isinstance(caller, Element):
                    Mnode = np.zeros((self.ndofs,self.ndofs))              # initialize to zeros
                    map = self.getIdx4DOFs(caller.getDofs(), local=True)   # get a map from element dogs to node dofs
                    Mnode[map[:, np.newaxis],map] = M                                     # fill in element matrix to nodal vector
                else:
                    Mnode = deepcopy(M)

                Mlocal = self._transform.m2l(Mnode, self)
            else:
                Mlocal = deepcopy(M)

            return Mlocal

        else:
            return self.lead.m2l(M, caller=caller)

    #
    # Mapping functions for Gauss-point values

    def _resetGaussPointMap(self, var=None):
        r"""
        Initialize or reset variable used for
        mapping gauss-point values to nodes
        """
        self._mapped_variable = var
        self._weighted_value  = 0.0
        self._weight          = 0.0

        if not self.is_lead:
            self.lead._resetGaussPointMap(var=var)

    def _addToMap(self, weight, weighted_value):
        """
        create nodal weighted sum of values and sum of weights
        for mapping of gauss-point values to nodes

        :param weight:
        :param weighted_value:
        """
        self._weight         += weight
        self._weighted_value += weighted_value

        if not self.is_lead:
            self.lead._addToMap(weight, weighted_value)

    def getMappedValue(self, var=None, ignore_lead=False):
        r"""
        Returns the mapped and weighted nodal value.

        :param var:  optional. If a string identifying the variable is given, this function will return zero i fthe current variable does not match
                     the once provided by the caller.
        :return:     the weighted nodal average for the requested variable.
        """
        if self.is_lead or ignore_lead:
            if var and var != self._mapped_variable:
                ans = 0.0
            else:
                if self._weight > 0.0:
                    ans = self._weighted_value / self._weight
                else:
                    ans = 0.0
            return ans
        else:
            return self.lead.getMappedValue(var)



import sys
import numpy as np
from copy import deepcopy


class Mesher():
    """

    """

    def __init__(self, model, *pts):
        self.model = model
        self.offset = np.array([0.0, 0.0])
        self.nodes    = []
        self.elements = []

    def shift(self, dx, dy):
        """
        :param dx: move patch to the right
        :param dy: move patch up
        """
        self.offset += np.array([dx, dy])

    def lineMesh(self, NeX, element_type, material, **kwargs):
        """
        nD mesher using line elements

        This is an abstract class - requires implementation in subclass

        :param int NeX: number of elements along the curve
        :param element_type: compatible element type
        :param material: a material object
        """
        msg = "** WARNING ** {}.{} not implemented".format(self.__class__.__name__, sys._getframe().f_code.co_name)
        raise NotImplementedError(msg)

    def quadMesh(self, NeX, NeY, element_type, material, **kwargs):
        """
        2D mesher using quadrilateral elements

        This is an abstract class - requires implementation in subclass

        :param int NeX: number of elements along the first axis
        :param int NeY: number of elements along the second axis
        :param element_type: compatible element type
        :param material: a material object
        """
        msg = "** WARNING ** {}.{} not implemented".format(self.__class__.__name__, sys._getframe().f_code.co_name)
        raise NotImplementedError(msg)

    def triangleMesh(self, NeX, NeY, element_type, material, **kwargs):
        """
        2D mesher using triangular elements

        This is an abstract class - requires implementation in subclass

        :param int NeX: number of elements along the first axis
        :param int NeY: number of elements along the second axis
        :param element_type: compatible element type
        :param material: a material object
        """
        msg = "** WARNING ** {}.{} not implemented".format(self.__class__.__name__, sys._getframe().f_code.co_name)
        raise NotImplementedError(msg)

    def brickMesh(self, NeX, NeY, NeZ, element_type, material, **kwargs):
        """
        Solid mesher using brick elements

        This is an abstract class - requires implementation in subclass

        :param int NeX: number of elements along the first axis
        :param int NeY: number of elements along the second axis
        :param int NeZ: number of elements along the third axis
        :param element_type: compatible element type
        :param material: a material object
        """
        msg = "** WARNING ** {}.{} not implemented".format(self.__class__.__name__, sys._getframe().f_code.co_name)
        raise NotImplementedError(msg)

    def tetMesh(self, NeX, NeY, NeZ, element_type, material, **kwargs):
        """
        Solid mesher using tetrahedral elements

        This is an abstract class - requires implementation in subclass

        :param int NeX: number of elements along the first axis
        :param int NeY: number of elements along the second axis
        :param int NeZ: number of elements along the third axis
        :param element_type: compatible element type
        :param material: a material object
        """
        msg = "** WARNING ** {}.{} not implemented".format(self.__class__.__name__, sys._getframe().f_code.co_name)
        raise NotImplementedError(msg)

    def tie(self, other, tol=1.0e-3):
        """
        Check nodes of **other** Mesher object for (near) identical
        location to any node in this Mesher object.  If overlap is identified, make the
        node in **other** a follower of the respective node in this Mesher.

        :param other: Mesher instance
        """
        if isinstance(other, Mesher):
            # tie the patches together
            for nd1 in self.nodes:
                for nd2 in other.nodes:
                    if np.linalg.norm(nd2.getPos() - nd1.getPos()) < tol:
                        nd2.make_follower(nd1)
        else:
            msg = "The tie(other) method requires a Mesher object as argument"
            raise TypeError(msg)




"""
====================
Example: truss01
====================

"""

from ...examples.Example import *

from ...domain.System import *
from ...domain.Node import *
from ...elements.Truss import *
from ...materials.FiberMaterial import *


class ExampleTruss01(Example):

    def docString(self):
        s = """
        Simple triangular truss. 
        
        The system is statically determined and allows for easy validation of 
        calculated deformation, reactions and internal forces.
        
        Author: Peter Mackenzie-Helnwein 
        """
        return s

    def problem(self):
        # initialize a system model
        B = 6.0 * 12
        H = 3.0 * 12
        params = {'E': 10., 'A': 1., 'nu': 0.0, 'fy': 1.e30}

        model = System()

        # create nodes
        nd0 = Node(0.0, 0.0)
        nd1 = Node(  B, 0.0)
        nd2 = Node(0.5*B, H)

        model.addNode(nd0, nd1, nd2)

        # create elements
        model.addElement(Truss(nd0, nd1, FiberMaterial(params)))  # bottom 1
        model.addElement(Truss(nd0, nd2, FiberMaterial(params)))  # up right diag 1
        model.addElement(Truss(nd1, nd2, FiberMaterial(params)))  # up left diag 1

        # define support(s)
        nd0.fixDOF('ux', 'uy')    # pin support left end
        nd1.fixDOF('uy')            # roller support right end

        # add loads
        # .. load only the upper nodes
        nd2.setLoad([0.0, -1.0], ('ux', 'uy'))

        # analyze the model
        model.solve()

        # write out report
        model.report()

        # create plots
        model.plot(factor=1., filename="truss01_deformed.png")
        model.beamValuePlot('f',filename="truss01_forces.png")

# %%
# Simple triangular truss.
#
# The system is statically determined and allows for easy validation of
# calculated deformation, reactions and internal forces.
#
# Author: Peter Mackenzie-Helnwein
#



import sys
import numpy as np

from .ShapeFunctions import ShapeFunctions
from .LineShapes import LineShapes

class QuadShapes(ShapeFunctions):
    r"""
    Shape functions or their :code:`*n-th` mixed derivative
    for the bi-unit-square (two-dimensional domain :math:`(s,t)\in[-1,+1]\times[-1,+1]`.

    Derivatives are with respect to the normalized coordinate and need to be scaled accordingly.

    """

    def __init__(self):
        super(QuadShapes, self).__init__(ShapeFunctions.QUADS)

    def shape(self, order, s, t, n=(0,0), serendipity=False):
        r'''

        :param order: polynomial order of the shape functions. Options: 0,1,2,
        :param s:    first coordinate of point of interest on interval [-1,1]
        :type s:  float or np.array
        :param t:   second coordinate location on interval [-1,1]
        :type t:  float or np.array
        :param n:     return n-th derivative of the requested shape function.
        :type n:  tuple or list

            Use :code:`n=(1,0)` for first derivative with respect to the first variable.
            Use :code:`n=(2,1)` for second derivative with respect to the first variable and
            first with respect to the second variable, i.e., a mixed derivative..

        :param serendipity: set to **True** for 8-node serendipity or **False** for bi-quadratic (9-node) quads.
        :type serendipity: bool

        :return: PHI: a list of shape function values
        '''
        ns, nt = n

        if (order == 0):
        #
        #   constant shape function
        #
            if (n == 0):
                PHI = np.array([1.0])
            else:
                PHI = np.array([0.0])

        elif (order == 1):
        #
        #   linear shape functions
        #
            if (ns == 0 and nt == 0):
                PHI = np.array([ 0.25 * (1-s) * (1-t),
                                 0.25 * (1+s) * (1-t),
                                 0.25 * (1+s) * (1+t),
                                 0.25 * (1-s) * (1+t) ])
            elif (ns == 1 and nt == 0):
                PHI = np.array([ -0.25 * (1-t),
                                  0.25 * (1-t),
                                  0.25 * (1+t),
                                 -0.25 * (1+t) ])
            elif (ns == 0 and nt == 1):
                PHI = np.array([ -0.25 * (1-s),
                                 -0.25 * (1+s),
                                  0.25 * (1+s),
                                  0.25 * (1-s) ])
            elif (ns == 1 and nt == 1):
                PHI = np.array([  0.25, -0.25, 0.25, -0.25 ])
            else:
                PHI = np.array([0.0, 0.0])

        elif (order == 2 and serendipity):
        #
        #   quadratic shape functions: 8-node quad
        #
            if (ns == 0 and nt == 0):
                PHI = np.array([ -0.25 * (1 - s) * (1 - t) * (1 + s + t),
                                 -0.25 * (1 + s) * (1 - t) * (1 - s + t),
                                 -0.25 * (1 + s) * (1 + t) * (1 - s - t),
                                 -0.25 * (1 - s) * (1 + t) * (1 + s - t),
                                  0.50 * (1 - s) * (1 + s) * (1 - t),
                                  0.50 * (1 + s) * (1 + t) * (1 - t),
                                  0.50 * (1 - s) * (1 + s) * (1 + t),
                                  0.50 * (1 - s) * (1 + t) * (1 - t) ])
            elif (ns == 1 and nt == 0):
                PHI = np.array([
                    -(0.25) * (-1 + t) * (2 * s + t), 
                    -(0.25) * (2 * s - t) * (-1 + t), 
                     0.25 * (1 + t) * (2 * s + t),
                     0.25 * (2 * s - t) * (1 + t),
                     s * (-1 + t),
                    -(0.50) * (-1 + t) * (1 + t),
                    -s * (1 + t),
                     0.50 * (-1 + t) * (1 + t)
                ])
            elif (ns == 0 and nt == 1):
                PHI = np.array([
                    -(0.25) * (-1 + s) * (s + 2 * t),
                    -(0.25) * (1 + s) * (s - 2 * t),
                     0.25 * (1 + s) * (s + 2 * t),
                     0.25 * (-1 + s) * (s - 2 * t),
                     0.50 * (-1 + s) * (1 + s),
                    -(1 + s) * t,
                    -(0.50) * (-1 + s) * (1 + s),
                    (-1 + s) * t
                ])
            elif (ns == 2 and nt == 0):
                PHI = np.array([
                    (1 - t)/2, (1 - t)/2, (1 + t)/2, (1 + t)/2, -1 + t, 0, -1 - t, 0
                ])
            elif (ns == 1 and nt == 1):
                PHI = np.array([
                    0.25 * (1 - 2 * s - 2 * t),
                    0.25 * (-1 - 2 * s + 2 * t),
                    0.25 * (1 + 2 * s + 2 * t),
                    0.25 * (-1 + 2 * s - 2 * t),
                    s,  -t,  -s,  t
                ])
            elif (ns == 0 and nt == 2):
                PHI = np.array([
                    (1 - s)/2., (1 + s)/2., (1 + s)/2., (1 - s)/2.,
                    0., -1. - s, 0., -1. + s
                ])
            elif (ns == 2 and nt == 1):
                PHI = np.array([
                    -(0.50), -(0.50), 0.50, 0.50, 1, 0, -1, 0
                ])
            elif (ns == 1 and nt == 2):
                PHI = np.array([
                    -(0.50), 0.50, 0.50, -(0.50), 0, -1, 0, 1
                ])
            else:
                PHI = np.array([0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0])

        elif (order == 2):
        #
        #   quadratic shape functions: 9-node quad
        #
            line = LineShapes()
            phi_s = line.shape(order=2, xi=0.5*(1+s), n=n[0])
            phi_t = line.shape(order=2, xi=0.5*(1+t), n=n[1])

            # make sure derivatives are on the bi-unit square
            if n[0]:
                phi_s *= (0.5) ** n[0]
            if n[1]:
                phi_t *= (0.5)**n[1]

            PHI = np.array(
                [
                    phi_s[0] * phi_t[0],  # bottom left
                    phi_s[1] * phi_t[0],  # bottom right
                    phi_s[1] * phi_t[1],  # top right
                    phi_s[0] * phi_t[1],  # top left
                    phi_s[2] * phi_t[0],  # bottom center
                    phi_s[1] * phi_t[2],  # right center
                    phi_s[2] * phi_t[1],  # top center
                    phi_s[0] * phi_t[2],  # left center
                    phi_s[2] * phi_t[2]   # center center
                ]
            )

        elif (order == 3):
        #
        #   cubic shape functions: using tensor multiplication of beam functions.
        #   these require rotations!
        #
            line = LineShapes()
            phi_s = line.shape(order=3, xi=0.5*(1+s), n=n[0], Le=2.)
            phi_t = line.shape(order=3, xi=0.5*(1+t), n=n[1], Le=2.)

            # make sure derivatives are on the bi-unit square
            if n[0]:
                phi_s *= (0.5) ** n[0]
            if n[1]:
                phi_t *= (0.5)**n[1]

            PHI = phi_s*phi_t  # this is an element-by element multiplication

        else:
            msg = "** WARNING ** {}.{} not implemented".format(self.__class__.__name__, sys._getframe().f_code.co_name)
            raise NotImplementedError(msg)

        return PHI



from .Material import *


class SectionMaterial(Material):
    r"""
    class: representing an integrated Section Material

    """

    def __init__(self, params={'E':1.0, 'A':1.0, 'I':1.0, 'nu':0.0, 'fy':1.0e30}):
        super().__init__(params = params)

        self._type = self.SECTION1D

        # make sure all necessary parameters exist
        if 'E' not in self.parameters:
            self.parameters['E']  = 1.0
        if 'A' not in self.parameters:
            self.parameters['A']  = 1.0
        if 'I' not in self.parameters:
            self.parameters['I']  = 1.0
        if 'nu' not in self.parameters:
            self.parameters['nu'] = 0.0
        if 'fy' not in self.parameters:
            self.parameters['fy'] = 1.0e30

        # initialize strain
        self.plastic_strain = 0.0
        self.setStrain({'axial':0.0, 'flexure':0.0})
        self.updateState()

    def updateState(self):
        EA = self.parameters['E']*self.parameters['A']
        EI = self.parameters['E']*self.parameters['I']

        if 'axial' in self.strain:
            force = EA * self.strain['axial']
        else:
            force = 0.0

        if 'flexure' in self.strain:
            moment = EI * self.strain['flexure']
        else:
            moment = 0.0

        self.stress = {'axial':force, 'flexure':moment}
        self.Et = {'ax-ax':EA, 'flx-flx':EI, 'ax-flx':0.0, 'flx-ax':0.0}




"""
Example
"""
from ...examples.Example import *

from ...domain.System import *
from ...domain.Node import *
from ...elements.Truss import *
from ...materials.FiberMaterial import *


class ExampleTruss04(Example):

    def docString(self):
        s = """
        3d truss example demonstrating large deformation analysis.
        
        Author: Tatsu 
        """
        return s

    def problem(self):

        # initialize a system model
        params = {'E': 2100., 'A': 1., 'nu': 0.0, 'fy': 1.e30}

        model = System()

        # create nodes
        H = 5
        nd1 = Node(0.0, 5.0, 0.0)
        nd2 = Node(9.5, 5.0, 0.0)
        nd3 = Node(0.0, 0.0, 0.0)
        nd4 = Node(9.5, 0.0, 0.0)
        nd5 = Node(5.5, 3.75, H)
        nd6 = Node(5.5, 1.25, H)

        nodeList = [nd1, nd2, nd3, nd4, nd5, nd6]
        model.addNode(*nodeList)


        # create elements
        model.addElement(Truss(nd1, nd5, FiberMaterial(params)))  # bottom 1
        model.addElement(Truss(nd1, nd6, FiberMaterial(params)))  # up right diag 1
        model.addElement(Truss(nd2, nd5, FiberMaterial(params)))  # up left diag 1
        model.addElement(Truss(nd3, nd6, FiberMaterial(params)))  # bottom 1
        model.addElement(Truss(nd4, nd5, FiberMaterial(params)))  # up right diag 1
        model.addElement(Truss(nd4, nd6, FiberMaterial(params)))  # up left diag 1
        model.addElement(Truss(nd5, nd6, FiberMaterial(params)))  # bottom 1


        # define support(s)
        translation_dofs = ('ux', 'uy', 'uz')
        for node in [nd1, nd2, nd3, nd4]:
            node.fixDOF(*translation_dofs)

        # add loads
        nd5.setLoad((-100.0,), ('uz',))

        # analyze the model
        model.solve()

        # write out report
        model.report()

        # create plots
        model.plot(factor=1.)



from copy import deepcopy

import numpy as np
import os
import sys
from .DrawElement import *
from ..recorder.Recorder import Recorder
from .Face2D import *
from .Face3D import *


class Element(DrawElement):
    r"""
    abstract class: representing a single generic element
    """

    COUNT = 0

    def __init__(self, nodes, material, label=None):
        r"""
        :param nodes: list of node pointers
        :param material: a :py:meth:`femedu.materials.Material` object
        """
        super(Element, self).__init__()

        self.ID = Element.COUNT # creates a unique ID for each element
        Element.COUNT += 1      # ensure the next call will create a unique ID as well

        self.label = label
        
        self.nodes    = nodes
        self.transforms = [ None for nd in self.nodes ]
        self.material = material
        self.dof_idx  = {}   # marked for removal

        self._requestDofs( tuple() )

        self.force    = 0.0
        self.Loads    = [ [] for i in range(len(nodes)) ]
        self.Forces   = []
        self.Kt       = []

        self.setRecorder(None)

        self.setLoadFactor(1.0)

    def __str__(self):
        if self.label:
            s = "{} ({}_{}): nodes ( ".format(self.label, self.__class__.__name__, self.ID)
        else:
            s = "{}_{}: nodes ( ".format(self.__class__.__name__, self.ID)
        for node in self.nodes:
            s += "{} ".format(node.getID())
        s += ")"
        s += "\n    material: {}".format(self.material.__class__.__name__)
        return s

    def __repr__(self):
        if self.label:
            fmt = "{}[{}](" + len(self.nodes)*"{}, " + "{})"
            ans = fmt.format(self.__class__.__name__, self.label,
                                    *[ node.getID() for node in self.nodes ],
                                    repr(self.material))
        else:
            fmt = "{}(" + len(self.nodes)*"{}, " + "{})"
            ans = fmt.format(self.__class__.__name__,
                                    *[ node.getID() for node in self.nodes ],
                                    repr(self.material))
        return ans

    def initialize(self, type=DrawElement.UNKNOWN, dofs=[] ):
        r"""
        :param type: define `DrawElement.TYPE` for plotting. See :py:meth:`femedu.elements.DrawElement` under **Inherited methods** for available options.
        :param dofs: list of dofs used by this element. Example: `["ux","uy"]` for 2d-displacements.
        """
        self.element_type = type

        if dofs:
            self._requestDofs(dofs)
        else:
            raise TypeError("mandatory list of dofs is missing for element {}".format(self))

        self.createFaces()

        nnodes = len(self.nodes)
        self.distributed_load = [ 0.0 for i in range(nnodes) ]

        self.reset_matrices()

    def reset_matrices(self):
        r"""
        (re-)initializes element stiffness matrix and element force
        """
        nnodes = len(self.nodes)
        ndofs  = len(self._dof_list)

        self.Forces   = [ np.zeros(ndofs) for k in range(len(self.nodes)) ]
        self.Kt       = [ [ np.zeros((ndofs,ndofs)) for k in range(nnodes) ] for m in range(nnodes) ]

        return (nnodes, ndofs)

    def resetLoads(self):
        r"""
        default implementation for resetting element loads.
        """
        for face in self.faces:
            face.setLoad(0.0, 0.0)

    def createFaces(self):

        self.faces = []

        if self.element_type == self.LINE:
            msg = "** WARNING ** {}.{} not implemented".format(self.__class__.__name__, sys._getframe().f_code.co_name)
            raise NotImplementedError(msg)

        elif self.element_type == self.CURVE :
            msg = "** WARNING ** {}.{} not implemented".format(self.__class__.__name__, sys._getframe().f_code.co_name)
            raise NotImplementedError(msg)

        elif self.element_type == self.TRIANGLE :
            nNds = len(self.nodes)
            if nNds == 3:
                self.faces.append(Face2D(f"{self.ID}.0", self.nodes[0],self.nodes[1]))
                self.faces.append(Face2D(f"{self.ID}.1", self.nodes[1],self.nodes[2]))
                self.faces.append(Face2D(f"{self.ID}.2", self.nodes[2],self.nodes[0]))
            elif nNds == 6:
                self.faces.append(Face2D(f"{self.ID}.0", self.nodes[0],self.nodes[3],self.nodes[1]))
                self.faces.append(Face2D(f"{self.ID}.1", self.nodes[1],self.nodes[4],self.nodes[2]))
                self.faces.append(Face2D(f"{self.ID}.2", self.nodes[2],self.nodes[5],self.nodes[0]))
            else:
                msg = "** WARNING ** {}.{} not implemented".format(self.__class__.__name__, sys._getframe().f_code.co_name)
                raise NotImplementedError(msg)

        elif self.element_type == self.TETRAHEDRON :
            nNds = len(self.nodes)
            if nNds == 4:
                self.faces.append(Face3D(f"{self.ID}.0", self.nodes[0],self.nodes[2],self.nodes[1]))
                self.faces.append(Face3D(f"{self.ID}.1", self.nodes[0],self.nodes[1],self.nodes[3]))
                self.faces.append(Face3D(f"{self.ID}.2", self.nodes[1],self.nodes[2],self.nodes[3]))
                self.faces.append(Face3D(f"{self.ID}.3", self.nodes[2],self.nodes[0],self.nodes[3]))
            elif nNds == 10 :
                self.faces.append(Face3D(f"{self.ID}.0", self.nodes[0],self.nodes[2],self.nodes[1],
                                                         self.nodes[6],self.nodes[5],self.nodes[4]))
                self.faces.append(Face3D(f"{self.ID}.1", self.nodes[0],self.nodes[1],self.nodes[3],
                                                         self.nodes[4],self.nodes[8],self.nodes[7]))
                self.faces.append(Face3D(f"{self.ID}.2", self.nodes[1],self.nodes[2],self.nodes[3],
                                                         self.nodes[5],self.nodes[9],self.nodes[8]))
                self.faces.append(Face3D(f"{self.ID}.3", self.nodes[2],self.nodes[0],self.nodes[3],
                                                         self.nodes[6],self.nodes[7],self.nodes[9]))
            else:
                msg = "** WARNING ** {}.{} not implemented".format(self.__class__.__name__, sys._getframe().f_code.co_name)
                raise NotImplementedError(msg)

        elif self.element_type == self.QUAD :
            nNds = len(self.nodes)
            if nNds == 4:
                self.faces.append(Face2D(f"{self.ID}.0", self.nodes[0],self.nodes[1]))
                self.faces.append(Face2D(f"{self.ID}.1", self.nodes[1],self.nodes[2]))
                self.faces.append(Face2D(f"{self.ID}.2", self.nodes[2],self.nodes[3]))
                self.faces.append(Face2D(f"{self.ID}.3", self.nodes[3],self.nodes[0]))
            elif nNds == 8 or nNds == 9 :
                self.faces.append(Face2D(f"{self.ID}.0", self.nodes[0],self.nodes[4],self.nodes[1]))
                self.faces.append(Face2D(f"{self.ID}.1", self.nodes[1],self.nodes[5],self.nodes[2]))
                self.faces.append(Face2D(f"{self.ID}.2", self.nodes[2],self.nodes[6],self.nodes[3]))
                self.faces.append(Face2D(f"{self.ID}.3", self.nodes[3],self.nodes[7],self.nodes[0]))
            else:
                msg = "** WARNING ** {}.{} not implemented".format(self.__class__.__name__, sys._getframe().f_code.co_name)
                raise NotImplementedError(msg)

        elif self.element_type == self.BRICK :
            nNds = len(self.nodes)
            if nNds == 8:
                self.faces.append(Face3D(f"{self.ID}.0", self.nodes[0],self.nodes[3],self.nodes[2],self.nodes[1]))
                self.faces.append(Face3D(f"{self.ID}.1", self.nodes[0],self.nodes[1],self.nodes[5],self.nodes[4]))
                self.faces.append(Face3D(f"{self.ID}.2", self.nodes[1],self.nodes[2],self.nodes[6],self.nodes[5]))
                self.faces.append(Face3D(f"{self.ID}.3", self.nodes[2],self.nodes[3],self.nodes[7],self.nodes[6]))
                self.faces.append(Face3D(f"{self.ID}.4", self.nodes[3],self.nodes[0],self.nodes[4],self.nodes[7]))
                self.faces.append(Face3D(f"{self.ID}.5", self.nodes[4],self.nodes[5],self.nodes[6],self.nodes[7]))
            elif nNds == 20 :
                self.faces.append(Face3D(f"{self.ID}.0", self.nodes[0],self.nodes[3],self.nodes[2],self.nodes[1],
                                                         self.nodes[11],self.nodes[10],self.nodes[9],self.nodes[8]))
                self.faces.append(Face3D(f"{self.ID}.1", self.nodes[0],self.nodes[1],self.nodes[5],self.nodes[4],
                                                         self.nodes[8],self.nodes[13],self.nodes[16],self.nodes[12]))
                self.faces.append(Face3D(f"{self.ID}.2", self.nodes[1],self.nodes[2],self.nodes[6],self.nodes[5],
                                                         self.nodes[9],self.nodes[14],self.nodes[17],self.nodes[13]))
                self.faces.append(Face3D(f"{self.ID}.3", self.nodes[2],self.nodes[3],self.nodes[7],self.nodes[6],
                                                         self.nodes[10],self.nodes[15],self.nodes[18],self.nodes[14]))
                self.faces.append(Face3D(f"{self.ID}.4", self.nodes[3],self.nodes[0],self.nodes[4],self.nodes[7],
                                                         self.nodes[11],self.nodes[12],self.nodes[19],self.nodes[15]))
                self.faces.append(Face3D(f"{self.ID}.5", self.nodes[4],self.nodes[5],self.nodes[6],self.nodes[7],
                                                         self.nodes[16],self.nodes[17],self.nodes[18],self.nodes[19]))
            elif nNds == 27 :
                self.faces.append(Face3D(f"{self.ID}.0", self.nodes[0],self.nodes[3],self.nodes[2],self.nodes[1],
                                            self.nodes[11],self.nodes[10],self.nodes[9],self.nodes[8],self.nodes[20]))
                self.faces.append(Face3D(f"{self.ID}.1", self.nodes[0],self.nodes[1],self.nodes[5],self.nodes[4],
                                            self.nodes[8],self.nodes[13],self.nodes[16],self.nodes[12],self.nodes[21]))
                self.faces.append(Face3D(f"{self.ID}.2", self.nodes[1],self.nodes[2],self.nodes[6],self.nodes[5],
                                            self.nodes[9],self.nodes[14],self.nodes[17],self.nodes[13],self.nodes[22]))
                self.faces.append(Face3D(f"{self.ID}.3", self.nodes[2],self.nodes[3],self.nodes[7],self.nodes[6],
                                            self.nodes[10],self.nodes[15],self.nodes[18],self.nodes[14],self.nodes[23]))
                self.faces.append(Face3D(f"{self.ID}.4", self.nodes[3],self.nodes[0],self.nodes[4],self.nodes[7],
                                            self.nodes[11],self.nodes[12],self.nodes[19],self.nodes[15],self.nodes[24]))
                self.faces.append(Face3D(f"{self.ID}.5", self.nodes[4],self.nodes[5],self.nodes[6],self.nodes[7],
                                            self.nodes[16],self.nodes[17],self.nodes[18],self.nodes[19],self.nodes[25]))
            else:
                msg = "** WARNING ** {}.{} not implemented".format(self.__class__.__name__, sys._getframe().f_code.co_name)
                raise NotImplementedError(msg)

        else:
            msg = "** WARNING ** {}.{} not implemented".format(self.__class__.__name__, sys._getframe().f_code.co_name)
            raise NotImplementedError(msg)

    def setSurfaceLoad(self, face_idx, pn, ps=0):
        r"""
        .. warning::

            This method needs to be implemented by every element that shall accept a surface load.

        :param face_ix: face index for the loaded face (integer starting at 0)
        :type face_idx: int
        :param pn: magnitude of distributed normal load per unit length. Tension on a surface is positive.
        :param ps: magnitude of distributed shear load per unit length. Positive shear rotates the element counter-clockwise.
        """
        msg = "** WARNING ** {}.{} not implemented".format(self.__class__.__name__, sys._getframe().f_code.co_name)
        raise NotImplementedError(msg)

    def addTransformation(self, T, local_nodes=[]):
        r"""
        Attach a transformation to any node of the element.

        If no **local_nodes** list is given or an empty list is handed to the function,
        the transformation, **T**, will be applied to all nodes in the element.

        A non-empty **local_nodes** list will apply the transformation to those local nodes listed in that list.
        Local nodes start at 0 and go to N-1, where N is the number of elements in this element.

        A transformation can be removed from a node by assigning :code:`T=None` as the transformation.
        """
        if local_nodes:
            for local_id in local_nodes:
                if local_id >= 0 and local_id < len(self.transforms):
                    T.registerClient(self)    # register this Element with the transformation
                    self.transforms[local_id] = T
        else:
            T.registerClient(self)            # register this Element with the transformation
            self.transforms = [ T for nd in self.nodes ]

    def getPos(self, node, **kwargs):
        r"""
        Use this function to get nodal displacements from inside your element implementation.

        This is a wrapper around the :py:meth:`Node.getPos(caller)` method that adds the caller information
        for the node.

        :param node: local node number for which displacements are requested.
        """
        return self.nodes[node].getPos(self, **kwargs)

    def getDisp(self, node, **kwargs):
        r"""
        Use this function to get nodal displacements from inside your element implementation.

        This is a wrapper around the :py:meth:`Node.getDisp(caller)` method that adds the caller information
        for the node.

        :param node: local node number for which displacements are requested.
        """
        return self.nodes[node].getDisp(caller=self, **kwargs)

    def getForce(self):
        r"""
        Request the internal force vector (stress driven force only; **no applied element loads**)

        :return:
        """
        self.updateState()

        # make sure forces are returned in each respective node's local coordinates
        forces = []
        for node, force in zip(self.nodes, self.Forces):
            forces.append(node.v2l(force, self))

        return forces

    def getLoad(self, apply_load_factor=False):
        r"""
        Requesting nodal forces generated by element loads, like

        * line loads on beams of frames
        * surface loads on plates or solids
        * body forces on solids.

        :param apply_load_factor: shall the global load factor be applied by the element.

        :return:  element load vector
        """
        #self.updateState()

        # .. applied element load (reference load)
        self.computeSurfaceLoads()

        if isinstance(self.Loads, (list,tuple,np.ndarray)):
            Loads = []
            for node, load in zip(self.nodes, self.Loads):
                if isinstance(load, np.ndarray):
                    Loads.append(node.v2l(load, self))
                else:
                    Loads.append(None)
            return Loads
        elif self.Loads:
            return self.Loads
        else:
            return [ None for k in self.nodes ]

    def computeSurfaceLoads(self):
        self.Loads = [ None for nd0 in self.nodes ]

    def getLabel(self):
        if self.label:
            return "{}".format(self.label)
        else:
            return self.getID()

    def getID(self):
        return "Elem_{}".format(self.ID)

    def getInternalForce(self, variable=''):
        msg = "** WARNING ** {}.{} not implemented".format(self.__class__.__name__, sys._getframe().f_code.co_name)
        raise NotImplementedError(msg)

    def getStress(self):
        self.updateState()
        return None

    def mapGaussPoints(self, var):
        r"""
        Initiate mapping of Gauss-point values to nodes.
        This method is an internal method and should not be called by the user.
        Calling that method explicitly will cause faulty nodal values.

        :param var: variable code for a variable to be mapped from Gauss-points to nodes
        """
        msg = "** WARNING ** {}.{} not implemented".format(self.__class__.__name__, sys._getframe().f_code.co_name)
        #raise NotImplementedError(msg)
        print(msg)

    def getStiffness(self):
        r"""
        :return: the current tangent stiffness matrix
        """
        self.updateState()

        KT = deepcopy(self.Kt)

        for i, ndI in enumerate(self.nodes):
            for j, ndJ in enumerate(self.nodes):
                #
                # this can be an expensive operation.
                # check if really needed and avoid identity operations:
                #
                if (ndI.hasTransform() or ndJ.hasTransform()):
                    KTij = KT[i][j]

                    # transform KTij
                    # ... as a sequence of vector-like transformations
                    KTij = KTij.T
                    if ndJ.hasTransform():
                        KTij = ndJ.m2l(KTij, self)

                    KTij = KTij.T
                    if ndI.hasTransform():
                        KTij = ndI.m2l(KTij, self)

                    KT[i][j] = KTij
        return KT

    def updateState(self):
        """
        """
        msg = "** WARNING ** {}.{} not implemented".format(self.__class__.__name__, sys._getframe().f_code.co_name)
        raise NotImplementedError(msg)

    def _requestDofs(self, dof_requests):
        r"""
        Helper function (internal use) to inform **all** nodes of this element about the needed/used
        degrees of freedom.

        **Remark**: if nodes of different type are to be used by the element, **DO NOT** use this method but
        implement your own overloaded initialization within the constructor of your element.

        :param dof_requests: list of dofs for a typical node in this element
        """
        self._dof_list = dof_requests
        for node in self.nodes:
            dof_idx = node.request(dof_requests, self)
            self.dof_idx[node] = dof_idx   # marked for removal

    def getDofs(self):
        r"""
        returns the dof-codes for this element in a list
        """
        return self._dof_list

    def setLoadFactor(self, lam):
        r"""
        Set the target load factor to **lam**

        .. warning::

            This method should not be called by the user.
            **USE** :code:`System.setLoadFactor(lam)` instead!

        The entered load pattern is considered a reference load,
        to be multiplied by a scalar load factor, :math:`\lambda`.

        If no load factor is set explicitly, a factor of 1.0 is assumed, i.e., the
        entire entered load is applied in full.
        """
        self.loadfactor = lam

    def setRecorder(self, recorder):
        if isinstance(recorder, Recorder):
            self.recorder = recorder
        else:
            self.recorder = None

    def recordThisStep(self, load_level):
        r"""
        record current state of the system
        """
        if self.recorder and self.recorder.isActive():
            data = {}
            for var in self.recorder.getVariables():
                value = self.getValue(var)[0]
                data[var] = value
            self.recorder.addData(data)

    def getValue(self, name, gp='all'):
        ans = []
        return ans
    def startRecorder(self):
        if self.recorder:
            self.recorder.enable()

    def pauseRecorder(self):
        if self.recorder:
            self.recorder.disable()

    def stopRecorder(self):
        if self.recorder:
            self.recorder.disable()

    def on_converged(self):
        r"""
        This method is called by the solver if convergence has been achieved.

        The element shall perform all necessary state updates,
        especially inform it's material instances about the necessary updates.
        """
        pass

    def revert(self):
        r"""
        This method is called by the solver if the current load step failed to converge
        and the system shall be returned to the last converged state.

        The element shall perform all necessary state updates,
        especially inform it's material instances about the necessary updates.
        """
        pass


if __name__ == "__main__":

    sys.path.insert(0, os.path.abspath(".."))

    from ..domain import Node
    from ..materials import Material

    # testing the Element class
    nd0 = Node(0.0, 0.0)
    nd0.index = 0
    nd1 = Node(3.0, 2.0)
    nd1.index = 1
    params = {'E':100, 'A':1.5, 'fy':1.0e20}
    mat = Material(params)
    elem = Element([nd0, nd1], mat)

    print(nd0)
    print(nd1)

    print("force =", elem.getAxialForce())
    print("nodal forces: ", *elem.getForce())
    print("element stiffness: ", elem.getStiffness())

    # change the nodal displacements
    nd0.setDisp(.1, .05)
    nd1.setDisp(.05, .2)

    print(nd0)
    print(nd1)

    print("force =", elem.getAxialForce())
    print("nodal forces: ", *elem.getForce())
    print("element stiffness: ", elem.getStiffness())





import numpy as np
from ..Element import *
from ...domain.Node import *

class Triangle(Element):
    """
    class: representing a single truss element
    """

    def __init__(self, node0, node1, node2, material, label=None):
        super().__init__((node0, node1, node2), material, label=label)
        self.element_type = DrawElement.TRIANGLE
        self.createFaces()

        if node0.getPos().size == 3:
            dof_list = ('ux','uy','uz')
            ndof = 3
        elif node0.getPos().size == 2:
            dof_list = ('ux','uy')
            ndof = 2
        else:
            raise TypeError("dimension of nodes must be 2 or 3")

        self._requestDofs(dof_list)

        self.distributed_load = [0.0, 0.0, 0.0]
        self.force    = 0.0
        self.Forces   = [ np.zeros(ndof) for k in range(len(self.nodes)) ]
        self.Kt       = [ [ np.zeros(ndof) for k in range(len(self.nodes)) ] for m in range(len(self.nodes)) ]

        # covariant base vectors (reference system)
        base1 = node1.getPos() - node0.getPos()
        base2 = node2.getPos() - node0.getPos()
        self.gcov = np.vstack((base1, base2))

        # metric (reference system)
        self.GIJ = self.gcov @ self.gcov.T

        # dual base vectors (reference system)
        self.gcont = np.linalg.inv(self.GIJ) @ self.gcov

        self.area = np.sqrt(np.linalg.det(self.GIJ)) / 2.0

    def __str__(self):
        s = super(Triangle, self).__str__()
        s += "\n    strain: xx={xx:.3e} yy={yy:.3e} xy={xy:.3e} zz={zz:.3e}".format(**self.material.getStrain())
        s += "\n    stress: xx={xx:.3e} yy={yy:.3e} xy={xy:.3e} zz={zz:.3e}".format(**self.material.getStress())
        if np.array(self.distributed_load).any():
            s += "\n    element forces added to node:"
            for i, P in enumerate(self.Loads):
                Pi = np.array(P) * self.loadfactor
                s += "\n        {}: {}".format(self.nodes[i].getID(), Pi)
        return s

    def setSurfaceLoad(self, face, pn, ps=0):
        """
        .. list-table::
            :header-rows: 1

            * - face ID
              - nodes defining that face
            * - 0
              - :py:obj:`node 0` to :py:obj:`node 1`
            * - 1
              - :py:obj:`node 1` to :py:obj:`node 2`
            * - 2
              - :py:obj:`node 2` to :py:obj:`node 0`


        :param face: face ID for the laoded face
        :param pn: magnitude of distributed normal load per area. Tension on a surface is positive.
        :param ps: magnitude of distributed shear load per area. Positive direction is defined as shown in the above table.
        """
        if face >= 0 and face <= 2:
            self.faces[face].setLoad(pn, ps)

    def resetLoads(self):
        super(Triangle, self).resetLoads()

    def updateState(self):

        node0 = self.nodes[0]
        node1 = self.nodes[1]
        node2 = self.nodes[2]

        Gs = self.gcont[0]
        Gt = self.gcont[1]
        Gu = -Gs - Gt

        # covariant base vectors (current system)

        gs = node1.getDeformedPos() - node0.getDeformedPos()
        gt = node2.getDeformedPos() - node0.getDeformedPos()
        gu = -gs - gt

        # metric (current system)
        gcov = np.vstack((gs, gt))
        gIJ = gcov @ gcov.T

        # deformation gradient
        F = np.outer(gs, Gs) + np.outer(gt, Gt)

        convective_strain = 0.5 * (gIJ - self.GIJ)
        almansi_strain = 0.5 * ( np.tensordot(F,F,((0,), (0,))) - np.eye(np.size(gs)) )

        # Dphi0_inv = np.linalg.inv(self.gcov.T)
        # val = Dphi0_inv.T @ convective_strain @ Dphi0_inv  # same as almansi_strain

        eps = almansi_strain

        # update the material state

        self.strain = {'xx':eps[0,0], 'yy':eps[1,1], 'xy':eps[0,1]+eps[1,0]}

        self.material.setStrain(self.strain)

        # 2nd Piola-Kirchhoff stress
        stress = self.material.getStress()

        S = np.array( [[stress['xx'],stress['xy']],[stress['xy'],stress['yy']]] )

        # 1st Piola-Kirchhoff stress
        P = F @ S

        # store stress for reporting
        self.stress = {'xx':P[0,0], 'xy':P[0,1], 'yx':P[1,0], 'yy':P[1,1]}

        # tractions
        ts = P @ Gs
        tt = P @ Gt
        tu = P @ Gu

        # initialize arrays
        gx = Gs[0] * gs + Gt[0] * gt
        gy = Gs[1] * gs + Gt[1] * gt

        # compute the kinematic matrices
        GI = (Gu, Gs, Gt)

        Bu = [Gu[0]*gx, Gu[1]*gy, Gu[1]*gx + Gu[0]*gy]
        Bs = [Gs[0]*gx, Gs[1]*gy, Gs[1]*gx + Gs[0]*gy]
        Bt = [Gt[0]*gx, Gt[1]*gy, Gt[1]*gx + Gt[0]*gy]

        BI = ( np.array(Bu), np.array(Bs), np.array(Bt) )

        # internal force
        self.Forces = [
            tu * self.area,
            ts * self.area,
            tt * self.area
            ]

        # tangent stiffness
        Ct = self.material.getStiffness() * self.area

        Kt = []
        One = np.eye(2, dtype=np.float64)
        for Gi, Bi in zip(GI, BI):
            Krow = []
            Ti = Gi @ S
            for Gj, Bj in  zip(GI, BI):
                GIJ = Ti @ Gj * self.area
                Krow.append( Bi.T @ Ct @ Bj + GIJ * One)
            Kt.append( Krow )

        self.Kt = Kt

        # .. applied element load (reference load)
        self.computeSurfaceLoads()

    def computeSurfaceLoads(self):
        """
        compute surface loads using faces

        This method should be called during :py:meth:`updateState()` by every
        element supporting surface loads

        """
        self.Loads = [ np.zeros_like(self.Forces[I]) for I in range(len(self.nodes)) ]

        for I, face in enumerate(self.faces):
            loads = face.computeNodalForces()

            # indexing
            J = I+1
            if J>2:
                J -= 3

            # add to element load vectors
            self.Loads[I] += loads[0]
            self.Loads[J] += loads[1]
            if loads.shape[0]>2:
                numNodes = len(self.nodes)
                if numNodes == 6:
                    K = I + 3
                elif numNodes == 8 or numNodes == 9:
                    K = I + 4
                else:
                    msg = "Force data provided from Face2D inconsistent with element data"
                    raise TypeError(msg)

                self.Loads[K] += loads[2]

    def getStress(self):
        return self.stress

    def mapGaussPoints(self, var):
        r"""
        Initiate mapping of Gauss-point values to nodes.
        This method is an internal method and should not be called by the user.
        Calling that method explicitly will cause faulty nodal values.

        :param var: variable code for a variable to be mapped from Gauss-points to nodes
        """
        # this element has a single gauss-point at s = t = u = 1/3

        stresses = ('sxx','syy','szz','sxy','syz','szx')
        membrane = ('nxx','nyy','nxy')
        strains  = ('epsxx','epsyy','epszz','epsxy','epsyz','epszx')

        value = 0.0

        if var.lower() in stresses:
            key = var[1:3].lower()
            #tensor = self.material.getStress()   # 2nd-Piola-Kirchhoff stress
            tensor = self.stress                  # 1st-Piola-Kirchhoff stress
            if key in tensor:
                thickness = self.material.getThickness()
                value = tensor[key]/thickness

        if var.lower() in membrane:
            key = var[1:3].lower()
            #tensor = self.material.getStress()   # 2nd-Piola-Kirchhoff stress
            tensor = self.stress                  # 1st-Piola-Kirchhoff stress
            if key in tensor:
                value = tensor[key]

        elif var.lower() in strains:
            key = var[1:3].lower()
            tensor = self.material.getStrain()  # Green-Lagrange Strain
            tensor = self.strain
            if key in tensor:
                value = tensor[key]

        wi     = self.area / 3.
        val_wi = value * wi

        for node in self.nodes:
            node._addToMap(wi, val_wi)



import matplotlib.pyplot as plt

from .Material import *


class FiberMaterial(Material):
    """
    class: representing a 1d Fiber Material

    """

    def __init__(self, params={'E':1.0, 'A':1.0, 'nu':0.0, 'fy':1.0e30}):
        super().__init__(params = params)

        self._type = self.FIBER

        # make sure all necessary parameters exist
        if 'E' not in self.parameters:
            self.parameters['E']  = 1.0
        if 'A' not in self.parameters:
            self.parameters['A']  = 1.0
        if 'nu' not in self.parameters:
            self.parameters['nu'] = 0.0
        if 'fy' not in self.parameters:
            self.parameters['fy'] = 1.0e30

        # initialize strain
        self.plastic_strain = 0.0
        self.setStrain({'xx':0.0})

    def getArea(self):
        return self.parameters['A']

    def updateState(self):
        """
        update state for a user provided axial strain value

        :param eps:  strain or strain tensor
        :return: n/a
        """
        # update stress state
        E  = self.parameters['E']
        fy = self.parameters['fy']

        eps = self.strain['xx']

        # elastic predictor
        self.sig = E * (eps - self.plastic_strain)
        self.Et = E

        # check yield condition
        f = np.abs(self.sig) - fy

        # plastic corrector as needed
        if f >= 0.0:
            depsP = np.sign(self.sig) * f/E
            self.sig -= E * depsP
            self.plastic_strain += depsP
            self.Et = 0.0

            print("material entering plastic state")

        #print(4*'{:12.8e}  '.format(eps, f, self.plastic_strain, self.sig ))
        self.stress = {'xx':self.sig, 'yy':0.0, 'zz':0.0, 'xy':0.0, 'xy':0.0, 'xy':0.0}

    def getStrain(self):
        return self.sig / self.parameters['E'] + self.plastic_strain


if __name__ == "__main__":
    # testing the Node class
    mat = FiberMaterial(params={'E':100.0, 'nu':0.0, 'fy':1.0})

    eps = 0.02 * np.sin( np.linspace(0, 2.*np.pi, 100) )

    sig = []
    Et = []
    for strain in eps:
        # update material strain state
        mat.setStrain(strain)
        #collect stress response
        sig.append(mat.getStress())
        #collect material tangent stiffness response
        Et.append(mat.getStiffness())

    fig, (ax1,ax2) = plt.subplots(2,1)

    ax1.plot(eps, sig, '-r', label='stress')
    ax1.grid(True)
    ax1.set_xlabel('strain $\\varepsilon$')
    ax1.set_ylabel('stress $\\sigma$')
    ax1.legend()

    ax2.plot(eps, Et, '-.b', label='tangent modulus')
    ax2.grid(True)
    ax2.set_xlabel('strain $\\varepsilon$')
    ax2.set_ylabel('tangent modulus $E_t$')
    ax2.legend()

    plt.show()



from .Faces import *

class Face3D(Faces):
    r"""
    Implementation representing one face of a 3D brick or tetrahedral element

    The face may be defined by

    * three nodes: liner interpolation of triangular face
    * four nodes:  linear interpolation of an 8-node brick.
    * six nodes:   quadratic interpolation of triangular face
    * eight nodes: quadratic interpolation of a 20-node brick.
    * nine nodes:  quadratic interpolation of a 27-node brick.

    .. note::

        The node order of 3D faces follows the node order of similar shaped 2D elements.

    """

    def __init__(self, id, *nds, **kwargs):
        super(Face3D, self).__init__(id, *nds, **kwargs)
        
    def initialize(self):
        r"""
        This method computes area and normals for :py:class:`Face3D`.
        
        It is called by the constructor of the :py:class:`Faces` class.
        """
        if self.dim < 3:
            msg = "{} requires 3D points - {} dimensions given".format(self.__class__.__name__, self.dim)
            raise TypeError(msg)

        super(Face3D, self).initialize()

    def isFace(self, X, N):
        r"""
        Implementation of the test function.

        :param X: position vector for a point
        :type X: np.array
        :param N: outward normal vector at **X**
        :type N: np.array
        :return: **True** if **X** and **N** match this face. **False** otherwise.
        """
        msg = "** WARNING ** {}.{} not implemented".format(self.__class__.__name__, sys._getframe().f_code.co_name)
        raise NotImplementedError(msg)





import math as m
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.collections import LineCollection
from matplotlib.colors import ListedColormap, BoundaryNorm
from matplotlib.patches import Arc, FancyArrow
import matplotlib.tri as tri

import sys

from .AbstractPlotter import *
from ..elements.Element import Element

class ElementPlotter(AbstractPlotter):
    r"""
    A plotter class that takes a list of nodes and elements from a finite element model
    and plots a deformed mesh, potentially shaded based on a value of a user-specified state
    variable.

    The information needed for plotting is obtained through standardized element functions.
    Default versions of those functions are implemented through the Element-class, but respective
    methods may be overloaded by user-implemented elements.
    """

    def __init__(self):
        super(ElementPlotter, self).__init__()

    def displacementPlot(self, factor=1.0, filename=None, modeshape=False, **kwargs):
        r"""
        Create a deformed system plot

        If **file** is given, store the plot to that file.
        Use proper file extensions to indicate the desired format (.png, .pdf)

        :param factor: displacement scaling factor
        :param file: filename (str)
        :param  kwargs: see list of optional parameters below.

        :keywords: optional parameters

            .. list-table::

                * - :py:obj:`show_bc`
                  - set to **True** to mark fixed DOFs
                * - :py:obj:`show_loads`
                  - set to **True** to plot loads
                * - :py:obj:`show_reactions`
                  - set to **True** to plot nodal reactions
                * - :py:obj:`principal`
                  - set to **True** to add principal stress orientation. Sets :py:obj:`factor=0`
                * - :py:obj:`pval`
                  - set to :py:obj:`stress` or :py:obj:`strain` to select tensor for principal value plot.
                    (defaults to :py:obj:`stress`)

        """

        #
        # setup for principal stress|strain plot
        #
        if 'principal' in kwargs:
            show_principal = kwargs['principal']
            if show_principal and 'pval' in kwargs:
                show_type = kwargs['pval']
                factor = 0.0
            else:
                show_type = 'stress'
        else:
            show_principal = False
            show_type = ''

        #
        # generic plot settings
        #
        if 'linewidth' in kwargs:
            lw1 = kwargs['linewidth']
            lw2 = 2*lw1
        else:
            lw1 = 1
            lw2 = 2

        if self.plot3D:
            fig = plt.figure(figsize=(10, 10))
            axs = fig.gca(projection='3d')

            # plot the undeformed elements
            for elem in self.elements:
                ans = elem.draw(factor=0.0, modeshape=False)
                if len(ans)>=3:
                    x = ans[0]
                    y = ans[1]
                    z = ans[2]
                    if x.size == y.size and x.size == z.size:
                        axs.plot(x, y, z, linewidth=1, linestyle='-', color='k')

            # plot the deformed elements
            if factor:
                for elem in self.elements:
                    ans = elem.draw(factor=factor, modeshape=modeshape)
                    if len(ans)>=3:
                        x = ans[0]
                        y = ans[1]
                        z = ans[2]
                        if x.size == y.size and x.size == z.size:
                            axs.plot(x, y, z, linewidth=2, linestyle='-', color='r')

            # plot orientation of principal stress|strain
            if show_principal:
                for elem in self.elements:
                    pass

            if self.reactions:
                self.addForces(axs, factor=factor)

            self.set_axes_equal(axs)

        else:
            if 'use_axis' in kwargs:
                fig, axs = kwargs['use_axis']
            else:
                fig, axs = plt.subplots()

            # plot the undeformed elements
            for elem in self.elements:
                ans = elem.draw(factor=0.0, modeshape=False)
                if len(ans)>=2:
                    x = ans[0]
                    y = ans[1]
                    if x.size == y.size:
                        axs.plot(x, y, linestyle='-', linewidth=lw1, color='k')

            # plot the deformed elements
            if factor:
                for elem in self.elements:
                    ans = elem.draw(factor=factor,modeshape=modeshape)
                    if len(ans)>=2:
                        x = ans[0]
                        y = ans[1]
                        if x.size == y.size:
                            if m.isclose(x[0], x[-1]) and m.isclose(y[0], y[-1]):
                                axs.fill(x, y, linestyle='-', linewidth=1, edgecolor='r', facecolor='grey', alpha=0.2)
                            axs.plot(x, y, linestyle='-', linewidth=lw2, color='r')

            # if self.reactions != []:
            #     self.addForces(axs)

            if 'show_bc' in kwargs and kwargs['show_bc']:
                d = 2.0  # should become a function of mesh dimension or image limits
                for node in self.nodes:
                    if node.isFixed('ux'):
                        pos = node.getPos()
                        x = [pos[0]-d, pos[0]+d]
                        y = [pos[1], pos[1]]
                        axs.plot(x,y,'-g',lw=1)
                    if node.isFixed('uy'):
                        pos = node.getPos()
                        x = [pos[0], pos[0]]
                        y = [pos[1]-d, pos[1]+d]
                        axs.plot(x,y,'-g',lw=1)
                    if node.isFixed('rz'):
                        pos = node.getPos()
                        axs.plot(pos[0],pos[1],'og',lw=1,ms=d)


            if 'show_loads' in kwargs and kwargs['show_loads'] and not modeshape:
                self.addForces(axs, loads=1, factor=factor)

            if 'show_reactions' in kwargs and kwargs['show_reactions'] and not modeshape:
                self.addForces(axs, reactions=1, factor=factor)

            if 'title' in kwargs:
                axs.set_title(kwargs['title'])
            else:
                axs.set_title(f"Deformed System (magnification={factor:.2f})")

            if 'use_axis' in kwargs:
                return

            axs.set_aspect('equal')
            axs.set_xmargin(0.20)
            axs.set_ymargin(0.20)
            axs.set_axis_off()

        if filename:
            plt.savefig(filename, bbox_inches='tight')
        plt.show()

    def valuePlot(self, variable_name='', factor=0.0, filename=None, gaussptvalue=False, **kwargs):
        """
        Create a plot using colors to identify magnitude of internal force.

        If **file** is given, store the plot to that file.
        Use proper file extensions to indicate the desired format (.png, .pdf)

        :param factor: True | **False**
        :param filename:  (str)
        :param gaussptvalue: False if variable_name is nodal dof, True if variable_name is a guass-point variable.
        """
        if 'cmap' in kwargs:
            cmap = kwargs['cmap']
        else:
            cmap = 'jet'

        if 'limits' in kwargs:
            limits = kwargs['limits']
            show_mesh = True
            if 'linewidth' not in kwargs:
                kwargs['linewidth'] = 0.125
        else:
            limits = [-1.e300, 1.e300]
            show_mesh = False

        if 'show_mesh' in kwargs:
            show_mesh = kwargs['show_mesh']

        if 'linewidth' not in kwargs:
            kwargs['linewidth'] = 0.125

        if self.plot3D:
            print("** WARNING ** {}.{} not implemented".format(self.__class__.__name__, sys._getframe().f_code.co_name))
            return

        else:
            fig, axs = plt.subplots()
            axs.set_aspect('equal')

            # build vertices
            verts = []
            values = []
            vert_ptr = {}
            for i, node in enumerate(self.nodes):
                verts.append(node.getPos())
                vert_ptr[node] = i
                if gaussptvalue:
                    val = node.getMappedValue(variable_name)
                else:
                    disps = node.getDisp(variable_name)

                    if not isinstance(disps, np.ndarray):
                        print(node)
                        print(disps)
                        raise

                    val = disps[0]

                if val < limits[0]: val = np.nan
                if val > limits[1]: val = np.nan
                values.append(val)

            values = np.array(values)
            verts  = np.array(verts)

            # build triangles
            triangles = []
            for elem in self.elements:
                if elem.isType(Element.TRIANGLE):
                    tri = [ vert_ptr[node] for node in [ elem.nodes[k] for k in [0,1,2] ] ]
                    triangles.append(tri)
                elif elem.isType(Element.QUAD):
                    tri = [ vert_ptr[node] for node in [ elem.nodes[k] for k in [0,1,2] ] ]
                    triangles.append(tri)
                    tri = [ vert_ptr[node] for node in [ elem.nodes[k] for k in [2,3,0] ] ]
                    triangles.append(tri)


            # adjust limits in case the range is too small
            if 'limits' in kwargs:
                limits = kwargs['limits']
                minVal = limits[0]
                maxVal = limits[1]
            else:
                minVal = values.min()
                maxVal = values.max()
                range = maxVal - minVal
                maxAbs = np.max([np.abs(minVal), np.abs(maxVal)])
                if maxAbs < 1.0e-6:
                    range = 1.0e-6
                elif range < 0.10 * maxAbs:
                    range = 0.10 * maxAbs
                minVal -= 0.05 * range
                maxVal += 0.05 * range

            # plot contours on undeformed elements
            tpc = axs.tripcolor(verts[:,0], verts[:,1], triangles, values,
                                vmin=minVal, vmax=maxVal,
                                shading='gouraud', edgecolors='k', cmap=cmap)
            fig.colorbar(tpc)

            if show_mesh:
                self.displacementPlot(factor=factor, use_axis=(fig, axs), **kwargs)

            if 'title' in kwargs:
                axs.set_title(kwargs['title'])
            else:
                axs.set_title(f"Contours of '{self.getLabel(variable_name)}'")

            #axs.set_xmargin(0.20)
            #axs.set_ymargin(0.20)
            axs.set_axis_off()

        if filename:
            plt.savefig(filename, bbox_inches='tight')
        plt.show()

    def xyPlot(self, X, Y, filename=None, show_arrows=False, **kwargs):
        """
        Create a x-y-plot

        If **filename** is given, store the plot to that file.
        Use proper file extensions to indicate the desired format (.png, .pdf)

        :param X: :py:class:`Record` holding x-values
        :param Y: :py:class:`list` or :py:class:`tuple` of :py:class:`Record` holding y-values
        :param str filename:
        :param str title: figure title (optional)
        :param str xlabel: x-axis label (optional)
        :param str ylabel: y-axis label (optional)
        """
        fig, axs = plt.subplots()

        xlbl, xi = X.getData()

        for Yi in Y:
            try:
                ylbl, yi = Yi.getData()
                axs.plot(xi, yi, ls='--', label=ylbl)
            except:
                print(f"warning: non-matching data record in XYplot()\n\tX:{X.label}\n\tY:{Yi.label}")

        x0 = X.getData()[1]
        axs.plot(x0, np.zeros_like(x0),'-k')
        axs.grid(True)
        axs.legend()

        if 'xlabel' in kwargs:
            axs.set_xlabel(kwargs['xlabel'])
        else:
            axs.set_xlabel(xlbl)

        if 'ylabel' in kwargs:
            axs.set_ylabel(kwargs['ylabel'])

        if 'title' in kwargs:
            axs.set_title(kwargs['title'])

        #axs.set_aspect('equal')
        #axs.set_xmargin(0.10)
        #axs.set_ymargin(0.10)

        if filename:
            plt.savefig(filename, bbox_inches='tight')
        plt.show()

    def beamValuePlot(self, variable_name='', factor=0.0, filename=None, show_arrows=False, **kwargs):
        """
        Create a deformed system plot

        If **file** is given, store the plot to that file.
        Use proper file extensions to indicate the desired format (.png, .pdf)

        :param variable_name: string code for variable
        :param factor: displacement scaling factor
        :param file: filename (str)
        """
        if not variable_name:
            print("No plotable variable defined")
            return

        if self.plot3D:
            raise NotImplementedError

            fig = plt.figure(figsize=(10, 10))
            axs = fig.gca(projection='3d')
            self.set_axes_equal(axs)

        else:
            fig, axs = plt.subplots()

            minX =  1.e20
            maxX = -1.e20
            minY =  1.e20
            maxY = -1.e20
            minV =  0.000
            maxV =  0.000

            # Find dimensions for scaling
            for elem in self.elements:
                ans = elem.draw(factor=0.0)
                if len(ans)>=2:
                    x = ans[0]
                    y = ans[1]
                    if x.size == y.size:
                        minX = np.min([minX, np.min(x)])
                        maxX = np.max([maxX, np.max(x)])
                        minY = np.min([minY, np.min(y)])
                        maxY = np.max([maxY, np.max(y)])
                (xsi, vals) = elem.getInternalForce(variable_name)
                if xsi.size != 0 and vals.size != 0:
                    minV = np.min([minV, np.min(vals)])
                    maxV = np.max([maxV, np.max(vals)])

            data_limits = (minV, maxV)

            # scaling factors
            if np.isclose(minX,maxX):
                maxX += 0.5
                minX -= 0.5
            if np.isclose(minY,maxY):
                maxY += 0.5
                minY -= 0.5
            if np.isclose(minV,maxV):
                maxV += 0.5
                minV -= 0.5

            Lx = maxX - minX
            Ly = maxY - minY
            Lv = maxV - minV

            scale = 0.150 * np.max([Lx,Ly]) / Lv

            # plot the elements
            for elem in self.elements:
                ans = elem.draw(factor=0.0)
                if len(ans)>=2:
                    x = ans[0]
                    y = ans[1]

                    if x.size == y.size and x.size>1:
                        xi = np.array([x[0],y[0]])
                        xj = np.array([x[-1],y[-1]])
                        lvec = xj - xi
                        ell  = np.linalg.norm(lvec)
                        nvec = lvec / ell
                        svec = np.array( [ -nvec[1], nvec[0] ] )

                        # variable plot
                        if elem.isType(Element.LINE) or elem.isType(Element.CURVE):
                            (xsi, vals) = elem.getInternalForce(variable_name)
                            if xsi.size > 0 and vals.size > 0:
                                vals *= scale

                                xv = xi + np.outer(xsi,  lvec)
                                vv = xv + np.outer(vals, svec)

                                axs.plot(vv[:,0], vv[:,1], '-g', lw=1)
                                self._arrow(axs, xv[0,0], xv[0,1], vv[0,0]- xv[0,0], vv[0,1]- xv[0,1], vals[0],show_point=show_arrows)
                                self._arrow(axs,xv[-1,0],xv[-1,1],vv[-1,0]-xv[-1,0],vv[-1,1]-xv[-1,1],vals[-1],show_point=show_arrows)

                        # plot system on top (!)
                        axs.plot(x, y, '-k', lw=2)

            # if self.reactions != []:
            #     self.addForces(axs)

            if variable_name.lower() == 'm' or variable_name.lower() == 'mz':
                # bending moment (in plane)
                axs.set_title('Bending Moment')
            elif variable_name.lower() == 't' or variable_name.lower() == 'mx':
                # bending moment (in plane)
                axs.set_title('Torque')
            elif variable_name.lower() == 'f' or variable_name.lower() == 'fx':
                # axial force
                axs.set_title('Axial Forces')
            elif variable_name.lower() == 'v' or variable_name.lower() == 'vy':
                # transverse shear (in-plane)
                axs.set_title('Shear Forces')
            else:
                # unknown force
                axs.set_title(f'{variable_name} Forces')

            y_limits = axs.get_ylim()
            msg = "data limits:  {:.2f} .. {:.2f}".format(*data_limits)
            axs.text(0.5*(minX+maxX), y_limits[0]-0.05*np.max([Ly,Lx]), msg, ha='center', va='top')

            axs.set_aspect('equal')
            axs.set_xmargin(0.10)
            axs.set_ymargin(0.10)
            axs.set_axis_off()

        if filename:
            plt.savefig(filename, bbox_inches='tight')
        plt.show()

    def _arrow(self,axs,x,y,dx,dy,val,show_point=True):

        if show_point:
            headWidth  = np.abs(val) * 0.20
            headLength = np.abs(val) * 0.50
        else:
            headWidth  = 0
            headLength = 0

        if val >= 0.0:
            fc = 'r'
            axs.arrow(x,y,dx,dy, fc=fc, ec=fc, lw=1,
                      length_includes_head=True,
                      head_width=headWidth, head_length=headLength)
        else:
            fc = 'b'
            axs.arrow(x+dx,y+dy,-dx,-dy, fc=fc, ec=fc, lw=1,
                      length_includes_head=True,
                      head_width=headWidth, head_length=headLength)

    def addForces(self, axs, loads=False, reactions=False, factor=0.0, dofs=('ux','uy','rz')):
        r"""
        add nodal forces to the plot shown in **axs**

        :param axs: axis on which to plot
        :param loads: set to True for loads added to the plot
        :param reactions: set to True for reactions added to the plot
        :param factor: deformation factor. default: 0.0
        :param dofs: dof mapping. default are 2D forces and in-plane moment
        """
        if loads and len(self.nodes) == len(self.loads):

            X=[]
            Y=[]
            Fx=[]
            Fy=[]

            for (node, force) in zip(self.nodes, self.loads):
                if np.linalg.norm(force) > 1.0e-3:
                    point = node.getDeformedPos(factor=factor)
                    X.append(point[0])
                    if point.size>1:
                        Y.append(point[1])
                    else:
                        Y.append(0.0)

                    if 'ux' in dofs and force.size>0:
                        Fx.append(force[0])
                    else:
                        Fx.append(0.0)

                    if 'uy' in dofs and force.size > 1:
                        Fy.append(force[1])
                    else:
                        Fy.append(0.0)

                    if 'rz' in dofs and force.size > 2 and np.abs(force[2])>1.e-3:
                        M = force[2]
                        w = 10
                        if M<0:
                            axs.add_artist( Arc((point[0],point[1]), w, w, angle=45., theta1=0.0, theta2=270.,
                                                color='blue') )
                            axs.add_artist( FancyArrow( point[0]+0.5*w*np.cos(np.radians(45.)),
                                                        point[1]+0.5*w*np.sin(np.radians(45.)),
                                                        2., -2., length_includes_head=True,
                                                        head_length=3, head_width=2, color='blue') )
                        else:
                            axs.add_artist( Arc((point[0],point[1]), w, w, angle=225., theta1=0.0, theta2=270.,
                                                color='green') )
                            axs.add_artist( FancyArrow( point[0]+0.5*w*np.cos(np.radians(135.)),
                                                        point[1]+0.5*w*np.sin(np.radians(135.)),
                                                        -2., -2., length_includes_head=True,
                                                        head_length=3, head_width=2, color='blue') )

            axs.quiver(X,Y, Fx, Fy, color='blue')

        if reactions and len(self.nodes) == len(self.reactions):

            X=[]
            Y=[]
            Fx=[]
            Fy=[]

            for (node, force) in zip(self.nodes, self.reactions):
                if np.linalg.norm(force) > 1.0e-3:
                    point = node.getDeformedPos(factor=factor)
                    X.append(point[0])
                    Y.append(point[1])
                    Fx.append(-force[0])
                    Fy.append(-force[1])

                    if force.size > 2 and np.abs(force[2])>1.e-3:
                        M = force[2]
                        w = 10
                        if M>0:
                            axs.add_artist( Arc((point[0],point[1]), w, w, angle=45., theta1=0.0, theta2=270.,
                                                color='green') )
                            axs.add_artist( FancyArrow( point[0]+0.5*w*np.cos(np.radians(45.)),
                                                        point[1]+0.5*w*np.sin(np.radians(45.)),
                                                        2., -2., length_includes_head=True,
                                                        head_length=3, head_width=2, color='green') )
                        else:
                            axs.add_artist( Arc((point[0],point[1]), w, w, angle=45., theta1=0.0, theta2=270.,
                                                color='green') )
                            axs.add_artist( FancyArrow( point[0]+0.5*w*np.cos(np.radians(135.)),
                                                        point[1]+0.5*w*np.sin(np.radians(135.)),
                                                        -2., -2., length_includes_head=True,
                                                        head_length=3, head_width=2, color='green') )

            axs.quiver(X,Y, Fx, Fy, color='green', pivot='tip')







import sys
import numpy as np
from mpl_toolkits import mplot3d
import matplotlib.pyplot as plt
from matplotlib.collections import LineCollection
from matplotlib.colors import ListedColormap, BoundaryNorm


class AbstractPlotter():
    r"""
    Defines all supported plot features.
    Implementation is required through derived classes.
    """

    def __init__(self):
        self.plot3D = False

        self.elements  = []  # element pointer list
        self.nodes     = []  # node pointer list
        self.factor    = 0.0 # displacement factor
        self.variable  = ''  # code for state variable to plot
        self.loads     = []  # list of nodal applied loads (from nodes and elements)
        self.reactions = []  # list of nodal reactions (from residual)

    def __str__(self):
        return "{}() object".format(self.__class__.__name__)

    def __repr__(self):
        return str(self)

    def setMesh(self, nodes, elems):
        r"""
        Link the nodes and elements so the ElementPlotter can get information from them.

        :param nodes:    list of node pointers
        :param elements: list of element pointer
        """
        self.nodes    = nodes
        self.elements = elems

    def setDisplacements(self, disp):
        r"""

        :param disp:
        """
        msg = "** WARNING ** {}.{} marked deprecated".format(self.__class__.__name__, sys._getframe().f_code.co_name)
        raise DeprecationWarning(msg)

    def setValues(self, vals):
        r"""

        :param vals:
        """
        msg = "** WARNING ** {}.{} marked deprecated".format(self.__class__.__name__, sys._getframe().f_code.co_name)
        raise DeprecationWarning(msg)

    def setReactions(self, R):
        r"""

        :param R: list or tuple of nodal force vectors
        """
        self.reactions = R

        # msg = "** WARNING ** {}.{} not implemented".format(self.__class__.__name__, sys._getframe().f_code.co_name)
        # raise NotImplementedError(msg)

    def setNodalLoads(self, P):
        r"""

        :param P: list or tuple of nodal force vectors
        """
        self.loads = P

        # msg = "** WARNING ** {}.{} not implemented".format(self.__class__.__name__, sys._getframe().f_code.co_name)
        # raise NotImplementedError(msg)

    def displacementPlot(self, factor=1.0, file=None):
        r"""
        Create a deformed system plot

        If **file** is given, store the plot to that file.
        Use proper file extensions to indicate the desired format (.png, .pdf)

        :param factor: displacement magnification factor
        :param file: filename (str)
        """
        msg = "** WARNING ** {}.{} not implemented".format(self.__class__.__name__, sys._getframe().f_code.co_name)
        raise NotImplementedError(msg)

    def valuePlot(self, variable_name='', deformed=False, file=None):
        r"""
        Create a plot using colors to identify magnitude of internal force.

        If **file** is given, store the plot to that file.
        Use proper file extensions to indicate the desired format (.png, .pdf)

        :param variable_name: string code for variable
        :param deformed: True | **False**
        :param file: filename (str)
        """
        msg = "** WARNING ** {}.{} not implemented".format(self.__class__.__name__, sys._getframe().f_code.co_name)
        raise NotImplementedError(msg)

    def beamValuePlot(self, variable_name='', factor=0.0, file=None):
        r"""
        Create a traditional beam value plot, i.e., moment and shear diagrams.

        If **file** is given, store the plot to that file.
        Use proper file extensions to indicate the desired format (.png, .pdf)

        :param variable_name: string code for variable
        :param factor: displacement scaling factor
        :param file: filename (str)
        """
        msg = "** WARNING ** {}.{} not implemented".format(self.__class__.__name__, sys._getframe().f_code.co_name)
        raise NotImplementedError(msg)

    def addForces(self, axs):
        r"""
        add nodal forces to the plot shown in **axs**

        :param axs: axis on which to plot
        """
        msg = "** WARNING ** {}.{} not implemented".format(self.__class__.__name__, sys._getframe().f_code.co_name)
        raise NotImplementedError(msg)

    def set_axes_equal(self, ax):
        r'''
        Make axes of 3D plot have equal scale so that spheres appear as spheres,
        cubes as cubes, etc..  This is one possible solution to Matplotlib's
        ax.set_aspect('equal') and ax.axis('equal') not working for 3D.

        :param ax: a matplotlib axis, e.g., as output from plt.gca().

        Source of this method from: https://stackoverflow.com/questions/13685386/matplotlib-equal-unit-length-with-equal-aspect-ratio-z-axis-is-not-equal-to
        '''

        x_limits = ax.get_xlim3d()
        y_limits = ax.get_ylim3d()
        z_limits = ax.get_zlim3d()

        x_range = abs(x_limits[1] - x_limits[0])
        x_middle = np.mean(x_limits)
        y_range = abs(y_limits[1] - y_limits[0])
        y_middle = np.mean(y_limits)
        z_range = abs(z_limits[1] - z_limits[0])
        z_middle = np.mean(z_limits)

        # The plot bounding box is a sphere in the sense of the infinity
        # norm, hence I call half the max range the plot radius.
        plot_radius = 0.5 * max([x_range, y_range, z_range])

        ax.set_xlim3d([x_middle - plot_radius, x_middle + plot_radius])
        ax.set_ylim3d([y_middle - plot_radius, y_middle + plot_radius])
        ax.set_zlim3d([z_middle - plot_radius, z_middle + plot_radius])


    def getLabel(self, variable):
        r"""
        Format known dof code strings as pretty LaTeX code

        :param variable: code string
        :return: string representing the variable using LaTeX
        """
        map = {
            'sxx': r'$\sigma_{xx}$',
            'syy': r'$\sigma_{yy}$',
            'szz': r'$\sigma_{zz}$',
            'sxy': r'$\sigma_{xy}$',
            'syz': r'$\sigma_{yz}$',
            'szx': r'$\sigma_{zx}$',
            'nxx': r'$h\sigma_{xx}$',
            'nyy': r'$h\sigma_{yy}$',
            'nxy': r'$h\sigma_{xy}$',
            'epsxx': r'$\varepsion_{xx}$',
            'epsyy': r'$\varepsion_{yy}$',
            'epszz': r'$\varepsion_{zz}$',
            'epsxy': r'$\gamma_{xy}$',
            'epsyz': r'$\gamma_{yz}$',
            'epszx': r'$\gamma_{zx}$',
            'qx': r'$q_{x}$',
            'qy': r'$q_{y}$',
            'qz': r'$q_{z}$',
            'ux': r'$u_{x}$',
            'uy': r'$u_{y}$',
            'uz': r'$u_{z}$',
            'rx': r'$\theta_{x}$',
            'ry': r'$\theta_{y}$',
            'rz': r'$\theta_{z}$',
            'Tx': r'$\partial T/\partial{x}$',
            'Ty': r'$\partial T/\partial{y}$',
            'Tz': r'$\partial T/\partial{z}$',
        }

        if variable in map:
            return map[variable]
        else:
            return variable



import numpy as np

from .FrameTransformation import *

class Frame2dTransformation(FrameTransformation):
    r"""
    A Transformation including displacement dofs and rotational dofs

    If vectors :math:`{\bf d}_1`  is given, the local coordinate frame :math:`\{{\bf g}_x,{\bf g}_y\}` is defined as follows.

    Normalized base

    .. math::
        {\bf g}_x = \frac{1}{||{\bf d}_1||} \: {\bf d}_1

    .. math::
        {\bf g}_y = {\bf e}_3 \times {\bf g}_x

    where :math:`{\bf e}_3` is the default out-of-plane unit vector.

    If instead an angle, :math:`\theta`, is given, the local coordinate frame is defined as

    .. math::
        {\bf g}_x = \cos\theta \: {\bf e}_x + \sin\theta \: {\bf e}_y

    .. math::
        {\bf g}_y = \sin\theta \: {\bf e}_x + \cos\theta \: {\bf e}_y

    where  :math:`\{{\bf e}_x,{\bf e}_y\}` is the global cartesian coordinate frame.

    :param dir1: in-plane vector, **dir1** == :math:`{\bf d}_1`, defining the local x-direction
    :param angle: rotation angle, :math:`\theta`, in *degrees*
    """

    def __init__(self, dir1=None, angle=None):

        dir1_is_vector = (isinstance(dir1, np.ndarray) or isinstance(dir1, list))

        if dir1_is_vector:
            if len(dir1)==2:
                dir2 = np.array([[0., -1.],[1., 0.]]) @ dir1 # vector perpendicular to the member axis
            else:
                msg = f"Wrong spatial dimension of dir1: {dir1}"
                raise TypeError(msg)

            dir2_is_vector = True

        if np.any(angle) and isinstance(angle,float):
            axis = np.radians(angle)
        else:
            axis = None

        super().__init__(dir1, dir2, axis)

        # reset vectors for transformation
        self.known_vectors = []
        self.addVector(['ux','uy'])
        # self.addVector(['rz'])    # rotation does not require transformation



from .Integration import Integration

class LineIntegration(Integration):
    """
    Provides integration points and weights for 1D integration
    """

    def __init__(self, order=2):
        super(LineIntegration, self).__init__(order=order, dimension=1)

        xi, w = self.gauss1D( nGP=(order + 2)//2 )
        self.xi = xi
        self.w  =  w
